# crs_funs.py    05Apr2020    crs
""" Our functions
Useful to us 
"""
import re 
from functools import cmp_to_key

from select_error import SelectError
from select_error_input import SelectErrorInput
###import select_trace

def anumcmp(a,b):
    """ Alphanumeric comaprison
    Treat as numeric comparison if a and b are numeric else alpha compare
     "12" > "3"
    :returns: 1,0,-1 if a>b, a==b, a<b    
     """
    if re.match(r'^\d+$', a) and re.match(r'^\d+$', b):
        cmp = int(a) - int(b)
        if cmp > 0:
            return 1
        if cmp < 0:
            return -1
        return 0
    if a == b:
        return 0
    
    return 1 if a > b else -1
 
def dotcmp(a,b):
    """ compare doted strings alphanumeric with sections within dots
        considered as numbers if both parts are strictly numeric
        e.g.  a.123.b is greater than a.33.b
    :a: first string
    :b: second string
        To facilitate comparing a123, a33,... a23b34c and a23c,...
        After spliting on ANY non word char, we find all alpha, digit strings
        We use re.findall because re.split does not support zero length
        re
    :returns: 1,0,-1 if a>b, a==b, a<b
    """
    a_ds = re.split(r'\W', a)
    b_ds = re.split(r'\W', b)
    a_dots = []
    for ad in a_ds:
        ads = re.findall(r'[^\W\d_]+|\d+', ad)
        if len(ads) > 0:
            a_dots.append(ads)
        else:
            a_dots.append(ad)   # No change

    b_dots = []
    for bd in b_ds:
        bds = re.findall(r'[^\W\d_]+|\d+', bd)
        if len(bds) > 0:
            b_dots.append(bds)
        else:
            b_dots.append(bd)   # No change
        
    for i in range(len(a_dots)):
        asec = a_dots[i]
        if i >= len(b_dots):
            return 1        # a longer
        bsec = b_dots[i]
        if not isinstance(asec, list):
            asec = [asec]
        if not isinstance(bsec, list):
            bsec = [bsec]
        for i in range(len(asec)):
            if i >= len(bsec):
                return 1
            
            ase = asec[i]
            bse = bsec[i]
            cmp = anumcmp(ase, bse)
            if cmp > 0:
                return 1
            
            if cmp < 0:
                return -1
        if len(asec) < len(bsec):
            return -1
            
    if len(b_dots) > len(a_dots):
        return -1
    
    return 1

def dotcmp_keyfunc():
    """ Return dotcmp as a key function for sorted
    """
    return cmp_to_key(dotcmp)

def dot_sorted(lst):
    """ Sort list of dotted text strings
    :lst: list of text strings
    :returns: sorted list 1.9.1 before 1.10.1
    """
    srt_list = sorted(lst, key=dotcmp_keyfunc())
    return srt_list

def str2bool(v):
    if re.match(r"^\d{2,}$", v):
        print(f"str2bool: {v} converted to 1")
        v = "1"
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    elif v == "0.0":
        ###select_trace.SlTrace.lg(f"Treat {v} as False")
        print(f"Treat {v} as False")
        return False
    else:
        #raise SelectError('Not a recognized Boolean value %s' % v)
        print('Not a recognized Boolean value %s - treat as False' % v)
        return False
    
def str2val(string, value_or_type):
    """ Convert string to value of type default
    :value_type: default value if string is None or variable type
            used as type expected
    :returns: converted value
            throws SelectErrorInput if conversion error
    """
    try:
        if isinstance(value_or_type, str) or value_or_type is str:
            return string
    
        if isinstance(value_or_type, bool) or value_or_type is bool:
            return str2bool(string)
            
        if isinstance(value_or_type, int) or value_or_type is int:
            return int(float(string))       # Treat floats as ints
        
        if isinstance(value_or_type, float) or value_or_type is float:
            return float(string)
    except Exception:
        ###SlTrace.lg(f"=>str2val input {string} error for type:{type(value)}")
        pass
    
    raise SelectErrorInput(f"str2val input {string} error for type:{type(value_or_type)}")

if __name__ == "__main__":
    from select_trace import SlTrace
    
    def test_it(lin):
        """ Test (actually exercise) on list
        """
        SlTrace.lg(f"List: {lin}")
        srt_list = dot_sorted(lin)
        SlTrace.lg(f"Sorted list: {srt_list}")
        SlTrace.lg()

    SlTrace.lg("Testing dotcmp - extended")
    dot_list = ["a10.b.c", "a2.b.c", "a3.b2.c", "a2.b1.c", "a2.b1.c2"]
    test_it(dot_list)
    dot_list = ["a10_b_c", "a2+b+c", "a3-b2-c", "a2=b1+c", "a2#b1#c2"]
    test_it(dot_list)
    dot_list = ["p1", "p10", "p3"]
    test_it(dot_list)
    dot_list = ["12a", "12b", "1a", "2", "10b", "11"]
    test_it(dot_list)
    
    SlTrace.lg("Testing dotcmp")
    dot_list = ["a.b.c", "1.2.3", "4.5.6", "7.8.9", "10.11.12", "13.14", "10.11", "4.5", "1", "2"]
    test_it(dot_list)
    
    dot2 = ["a1.b1.c1", "a.10", "a.2", "a.3", "a.21.c", "b.23.c", "b.100.c.d", "a.100.1"]
    test_it(dot2)
    dot2 = ["a1.b1.c1", "a.10", "a.10.1", "a.10.10", "a.9.10", "a.9", "a.2", "a.3", "a.21.c", "b.23.c", "b.100.c.d", "a.100.1"]
    test_it(dot2)
    
    SlTrace.lg("End of test")
    



#wx_turtle_braille.py       07Oct2025  crs ck if main_mod has __file__
#                           02Nov2023  crs from turtle_braille.property
#                           27Feb2023  crs from turtle_braille.py
#                           21Feb2023  crs  From turtle_braille.py
#                           16Apr2022  crs  Author
"""
# Using tk.Canvas scanning
Turtle augmented with braille graphics output
__main__ : main process: tkinter display support via done()

else: multiprocessing process:
    Create braille output 
    Create/control wxPython display AudioDrawWindow
    
"""

import os
import datetime
import sys
import re
import select
from threading import Thread
import subprocess
import tkinter as tk
import turtle as tur
from turtle import *

from wx_tk_rpc_host import TkRPCHost
from select_trace import SlTrace
from wx_canvas_grid import CanvasGrid
from wx_braille_cell_list import BrailleCellList
"""
External functions 
Some day may model after turtle's _make_global_funcs
"""
SlTrace.clearFlags()    # Start quiet
tkh = None
root = tk.Tk()
root.withdraw()     # Hide tk window
canvas = None
pdisplay = None

def setup_main(title=None, port=None):
    global tkh
    global canvas
    global pdisplay
    
    if canvas is None:
        canvas = tur.getcanvas()
    cg = CanvasGrid(base=canvas)
    cells = cg.get_cell_specs()
    SlTrace.lg(f"\nsetup_main: cells:{cells}", "cell_specs")
    src_file = __file__     # To be replaced with src file
    if '__main__' in sys.modules:        
        main_mod = sys.modules['__main__']
        if hasattr(main_mod, '__file__'):
            src_file = main_mod.__file__
    base_src_file = os.path.basename(src_file)
    id_title = f"File:{base_src_file}"    
    current_time = str(datetime.datetime.now())
    mt = re.match(r'(.*):\d+\.\d+$', current_time)
    if mt is not None:
        current_time = mt.group(1)  # Ignore seconds
    username = os.getlogin()
    id_title += f"  Date:{current_time}"
    if username is not None and username != "":
        id_title += f"  User:{username}"

    #root.withdraw()
    if tkh is None:
        tkh = TkRPCHost(canvas_grid=cg, root=root,host_port=port)
    src_dir = os.path.dirname(__file__)
    id_title = id_title.replace(" ", "_")
    title = id_title.replace(" ", "_")
    SlTrace.lg(f"setup_main: {id_title = }")
    pdisplay = subprocess.Popen(f"py wx_display_main.py"
                                f" --id_title {id_title}"
                                f" --title {title}"
                                f" --host_port={tkh.host_port}"
                                f" --src_file={src_file}"
                                 " --subprocess",
                    cwd=src_dir,
                    shell=True)
    check_display()
    tkh.wait_for_user()     # Wait till setup, possibly snapshot is done

n_check = 0
def mainloop(title=None, file=None, port=None):
    """ tk's mainloop
    :port: host socket port
    """

    global tkh
    global canvas
    
    if pdisplay is None:
        setup_main(title=title, port=port)
    tur.mainloop()
             
def check_display():
    """ Check if display process exited
    Recheck after delay
    """
    global n_check
    if pdisplay is None:
        return              # Start checking when launched
    
    n_check += 1
    rc = pdisplay.poll()
    #SlTrace.lg(f"check_display: {n_check}")
    if rc != None:
        SlTrace.lg(f"Subprocess exited with rc:{rc}")
        root.destroy()
        SlTrace.onexit()    # Close log
        os._exit(0)     # Stop all processes
        return
    tur.ontimer(check_display, 1000)
    #tur.ontimer(check_display, 10)
        
#tur.mainloop()
#sys.exit(0)

snap_inc = 0        # Augment port
def snapshot(title=None, port=None):
    """ Create a TurtleBraille window with a "snapshot" of current turtle display
    :title: Title description default: generated
    """
    SlTrace.lg(f"snapshot title={title} port={port}")
    if pdisplay is None:
        setup_main(title="Setup for snapshots", port=port)                # Create first window with current display
    tkh.snapshot(title=title)

        
def done(port=None):
    mainloop(title="From turtle pgm", port=port)

if __name__ ==  '__main__':
    colors = ["red","orange","yellow","green"]
    n = 0
    for colr in colors:
        n += 1
        width(40)
        color(colr)
        forward(200)
        right(90)
        #if n in [1,2,3,4]:
        if n in []:
            snapshot(f"{n}: {colr}")
    done()		    # Complete drawings
    
    


# select_trace.py
"""
trace/logging package
modeled after smTrace.java

Debug Tracing with logging support
"""
import re
import os
import atexit

from datetime import datetime
import sys
import traceback
import difflib
import tkinter as tk

from crs_funs import str2bool, str2val
from select_error import SelectError
from select_report import SelectReport
from java_properties import JavaProperties

"""
Facilitate execution tracing / logging of execution.
 
@author raysmith
"""

class TraceError(Exception):
    """Base class for exceptions in this module."""
    pass

class SlTrace:
    traceFlagPrefix = "traceFlag"

    """
    Logging Support
    """
    propName = None     # Properties file name None - use script base name
    newExt = None       # if not None => output will have newExt extension
    update = True       # True => write out updated properties at end (save_propfile)     
    propPathSaved = None   # Actual saved path, if saved
    logName = None # Logfile name
    logExt = "sllog" # Logfile extension (without ".")
    started = False     # Set True when logging is started
    closed = False      # Set True when log file is closed
    logWriter = None
    logToScreen = True # true - log, additionally to STDOUT
    stdOutHasTs = False # true - timestamp prefix on STDOUT
    lgsString = "" # Staging area for lgs, lgln
    traceObj = None
    decpl = 0       # Decimal places in time stamp
    defaultProps = None # program properties
    loadedProps = None  # loaded properties
    traceAll = False # For "all" trace
    traceFlags = {} # tracing flag/levels
    recTraceFlags = {} # recorded
    mem_used = 0        # Memory used as of last getMemory call
    mem_used_change = 0 # Memory change as of last getMemory call   
    runningJob = True
    properties_diff_sn = None       # Most recently used snapshot
    mw = None                       # Master window widget, if one
    mw_standalone = False           # Set True if we're responsible for cleanup
    all_quiet = False               # Eliminate all trace flagged output
    start_time = datetime.now()     # Starting time
    
    @classmethod
    def getDefaultProps(cls):
        """ Return properties access
        """
        if cls.defaultProps is None:
            cls.setupLogging()
            cls.setProps(logtoScreen=False)
        return cls.defaultProps
    
    @classmethod
    def setup(cls):
        dummy = 0
    
    @classmethod
    def now(cls):
        """ Shorthand for datetime access
        """
        return datetime.now()
    
    @classmethod
    def set_start_time(cls, start_time=None):
        """ Set beginning for duration
        :start_time: (datetime) starting time
                    default: now
        """
        if start_time is None:
            start_time = cls.now()
        cls.start_time = start_time
    
    @classmethod
    def time_since(cls, tend=None, tbeg=None):
        """ Set beginning for duration
        :tend: (datetime) end of duration
                    default: now
        :tbeg: (datetime) beginning of duration
                    default: start_time default: pgm start
        :returns: duration in seconds (float)
        """
        if tend is None:
            tend = cls.now()
        if tbeg is None:
            tbeg = cls.start_time
            
        tdelta = tend-tbeg
        tdiff_sec = tdelta.total_seconds()
        return tdiff_sec
            
    @classmethod
    def tr(cls, flag, level=1):
    
        cls.setup()        # Insure connection
        return cls.tr1(flag, level=level)

    @classmethod
    def clearFlags(cls):
        """ Clear all trace flags
        """
        cls.setupLogging()          # Insure setup
        for flag in cls.traceFlags:
            val = cls.traceFlags[flag] 
            if isinstance(val, bool):
                cls.traceFlags[flag] = False
            elif isinstance(val,int):
                cls.traceFlags[flag] = 0
            else:
                cls.traceFlags[flag] = 0
        cls.traceAll = 0            # For "all" trace

    @classmethod
    def setAllQuiet(cls, value=True):
        cls.all_quiet = value
        print(f"Setting all_quiet={cls.all_quiet}")
    
    @classmethod
    def diff(cls, file1, file2, req=None, req_match=None):
        """ Compare files
        :file1:  first
        :file2: second
        :req: required string only print lines including this string
                default: no required string
        req_match:  reqired re.match
                default: no required match
        """
        try:
            lines1 = open(file1).readlines()
        except:
            raise SelectError(f"diff missing file {file1}")
        
        try:
            lines2 = open(file2).readlines()
        except:
            raise SelectError(f"diff missing file {file2}")
        
        for line in difflib.unified_diff(lines1, lines2,
                                          fromfile='file1', tofile='file2', lineterm='\n', n=0):
            for prefix in ('---', '+++', '@@'):
                if line.startswith(prefix):
                    break
            else:
                if ((req is not None and line.find(req) >= 0)
                        or (req_match is not None and re.match(req_match, line))):
                    cls.lg(line)        

    @classmethod
    def diff_prop_file(cls):
        """ Prints change in most recent properties file
        """
        prop_file_beg = SlTrace.getPropPath()
        prop_file_new = SlTrace.getPropPathSaved()
        cls.diff(prop_file_beg, prop_file_new, ".name")

    
    """
    Setup trace flags from string of the form flag1[=value][,flagN=valueN]* Flags
    are case-insensitive and must not contain ",". Values are optional and
    default to 1 if not present.
    """
    @classmethod
    def setFlags(cls, settings):
        cls.setup() # Insure access
        pat_flag_val = re.compile(r"(\w+)(=(\w+),*)?")
        sets = re.findall(pat_flag_val, settings)
        for setting in sets:
            flag, value = setting[0], setting[2]
            if value is None or value == '':
                value = 1        # Defaule value if no =
            cls.setLevel(flag, value)


    @classmethod
    def setLogExt(cls, logExt):
        """ Setup Logging file default extension This extension is used if no extension
        is provided by user
        """
        cls.logExt = logExt


    @classmethod
    def setLogName(cls, logName):
        """ Setup Logging file name Name without extension is appended with
            "_YYYYMMDD_HHMMSS.tlog"
        """
        cls.logName = logName
        if cls.defaultProps is None:
            propName = cls.propName
            if propName is None:
                propName = os.path.basename(logName)
                propName = os.path.splitext(propName)[0]
            cls.setProps(propName)


    @classmethod
    def setLogToStd(cls, on):
        """ Set logging to stdout
        :on: output to screen
            default: log to screen
        """
        if on is None:
            on = True
        cls.logToScreen = on

    @classmethod
    def setLogStdTs(cls, on=None):
        """ Set stdout log to have timestamp prefix
        """
        if on is None:
            on = True
        cls.stdOutHasTs = on


    @classmethod
    def setupLogging(cls, logName=None,propName=None, dp=None,
                     logToScreen=None,
                     stdOutHasTs=None):
        """
        Setup writing to log file ( via lg(string)) if not already setup
        :logName: name or prefix to log file
        :propName: name for properties file
                default: generated
        :dp: number of decimal places in timestamp
                Iff present, modify current setting
        :logToScreen: put log to screen in addition to log file
            default: output goes only to file
            iff present, modify current setting
        :stdOutHasTs: put timestamp on STDOUT (screen)
            default: no timestamp on screen
            if present, modify current setting
        """

        if dp is not None:
            cls.decpl = dp          # Set/change default ts decimal places
        if logToScreen is not None:
            cls.logToScreen = logToScreen
        if stdOutHasTs is not None:
            cls.stdOutHasTs = stdOutHasTs
        if propName is not None:
            cls.propName = propName
        if cls.started is True:
            return cls.logWriter
        
        if logName is None:
            logName = cls.logName
    
        if logName is None:
            script_name = os.path.basename(sys.argv[0])
            script_name = os.path.splitext(script_name)[0]
            logName = script_name
        if not os.path.isabs(logName):
            cwd = os.getcwd()
            dir_base = os.path.basename(cwd)
            if dir_base == "src":
                log_dir = os.path.join("..", "log")  # Place in sister directory
                log_dir = os.path.abspath(log_dir)
            else:
                log_dir = "log"
            logName = os.path.join(log_dir, logName)
        if "." not in logName:
            if logName.endswith("_"):
                pass
            else:
                logName += "_"
            ts = cls.getTs()
            logName += ts
            cls.logName = logName
            logName += "."
            logName += cls.logExt  # Default extension
            cls.logName = logName
        cls.logName = logName    

        base_file = os.path.abspath(cls.logName)
        directory = os.path.dirname(base_file)
        if not os.path.exists(directory):
            try:
                os.mkdir(directory)
            except:
                raise TraceError("Can't create logging directory %s"
                                     % directory)
                sys.exit(1)
        
            print("Logging Directory %s  created\n"
                  % os.path.abspath(directory))

        fw = None
        try:
            abs_logName = os.path.abspath(cls.logName)
            fw = open(abs_logName, "w", encoding="utf-8")
            cls.logWriter = fw
        except IOError as e:
            print("Can't open logWriter %s - %s" % (abs_logName, e))
            sys.exit(1)
        cls.started = True
        atexit.register(cls.onexit)        # close down at end
        cls.lg("Creating Log File Name: %s" % abs_logName)
        cls.setProps(cls.propName, logtoScreen=False)         # Setup properties file

        return fw

    @classmethod
    def set_mw(cls, mw):
        """ Set tk to support gui
        :tk:  master widget
        """
        cls.mw = mw

    @classmethod
    def select_all(cls, level=1):
        """ Trace all known flags at level 1
        """
        for flag in cls.getAllTraceFlags():
            cls.setTraceFlag(flag, level)

    @classmethod
    def remove_non_ascii(cls, text, repl=""):
        """ Remove non_ascii
        :text: original text
        :repl: replacement for non-ascii characters
        """
        msg_str = re.sub(r'[^\x00-\x7f]+',repl, text)
        return msg_str


    @classmethod
    def select_none(cls, level=0):
        """ Trace all known flags at level 0
        """
        for flag in cls.getAllTraceFlags():
            cls.setTraceFlag(flag, level)


    @classmethod
    def lg(cls, msg="", trace_flag=None, level=1, to_stdout=None,
           dp=None, replace_non_ascii="???"):
        """
        Log string to file, and optionally to console STDOUT display is based on
        trace info
        @param msg
                   - message to output
        @param trace_flag
                   - when to trace - None - always
        @param level
                   - level to trace - default 1
        @param to_stdout
                   - put output to STDOUT (screen) in addition to log file
                     default: use setup value
        @param dp decimal points in timestamp seconds
                default: use pgm default
        @replace_non_ascii: replacement for all non-ascii characters
                        default: ???, None: no replacement
        @throws IOException
       
        @throws IOException
        """
        if cls.all_quiet and trace_flag is not None:
            return
        
        if not SlTrace.trace(trace_flag, level):
            return

        try:
            cls.setupLogging()
        except:
            print("IOException in lg setupLogging")
            return
        if dp is None:
            if SlTrace.trace("decpl"):
                dp = 3          # Set loging ts decimal places
            else:
                dp = cls.decpl 
        ts = cls.getTs(dp=dp)
        if to_stdout is None:
            to_stdout = cls.logToScreen
        if to_stdout:
            prefix = ""
            if cls.stdOutHasTs or SlTrace.trace("stdOutHasTs"):
                prefix += " " + ts
            try:
                msg_str = prefix + " " + str(msg)
                if replace_non_ascii is not None:
                    msg_str = cls.remove_non_ascii(msg_str, replace_non_ascii)
                print(msg_str)
                sys.stdout.flush()   # Force output
            except:
                try:
                    print("Unexpected error in", str(msg_str))
                except:
                    print("Unexpected error: - can't print msg_str")
                    return
                
                info = sys.exc_info()[0]
                try:
                    print("Unexpected error:", info, "\n")
                except:
                    print("Can't print info\n")
                pass
            
        if cls.closed:
            return
        
        if cls.logWriter is None:
            print("Can't write to log file")
            return

        try:
            out_str = ""
            if not cls.trace("no_ts"):
                out_str += " "+ts
            out_str += msg
            if replace_non_ascii is not None:
                out_str = cls.remove_non_ascii(out_str, replace_non_ascii)
            print(out_str, file=cls.logWriter)
            cls.logWriter.flush()    # Force output
        except IOError as e:
            print("IOError in lg write %s" % str(e))
        except:
            print("Unexpected error:", sys.exc_info()[0])   


    """
    Append string to log to be logged via next lgln() call
    
    @throws IOException
    """
    @classmethod
    def lgs(cls, msg):
        if cls.lgsString is None:
            lgsString = msg
        else:
            lgsString += msg

    """
    Output staged string to log
    
    @throws IOException
    """
    @classmethod
    def lgln(cls):
        if cls.lgsString is None:
            cls.lgsString = ""

        cls.lg(cls.lgsString)
        cls.lgsString = None # Flush pending


    @classmethod
    def save_propfile(cls):
        """ Save properties file - snapshot
            Can be called repeatedly, also called via atexit
        """
        print("Executing save_propfile")
        if cls.trace("prop_change"):
            cls.properties_change_print()
        if not cls.update:
            print("Not writing updated properties file")
            return
        
        try:
            if cls.defaultProps is not None:
                abs_propName = cls.defaultProps.get_path()
                if cls.newExt is not None:
                    m = re.match(r'(^.*)\.[^.]*$', abs_propName)
                    if m:
                        abs_propName = m[1] + "." + cls.newExt
                tsfmt = "%Y%m%d_%H%M%S"
                ts = datetime.now().strftime(tsfmt)
                title = f" {abs_propName}  {ts}"
                cls.lg(f"Saving properties file {title}")
                cls.propPathSaved = abs_propName
                outf = open(abs_propName, "w")
                list_props = cls.trace("list_props")
                cls.defaultProps.store(outf, title,
                        list_props = list_props)
                cls.defaultProps = None     # Flag as no longer available
                outf.close()
        except IOError as e:
            tbstr = traceback.extract_stack()
            cls.lg("Save propName %s store failed %s - %s"
                    % (abs_propName, tbstr, str(e)))

    @classmethod
    def snapshot_properties(cls, sn=None,
                                req=None, req_not=False,
                                req_match=None,
                                req_match_not=False):
        """ Properties snapshot, selected by req string and or req_match
        :sn: properties state
            default: current properties state
        :req: only keys which have this string
        :req_not: only keys that don't have this string
        :req_match: only keys which match this regular expression
        :req_match_not: only keys which don't match this regex
                
        :returns: (JavaProperties(only in sn1),
                  JavaProperties(sn1, sn1,sn2 differ),
                  JavaProperties(sn2, sn1,sn2 differ),
                  JavaProperties (only in sn2)) tuple
        """
        if sn is None:
            sn = cls.getDefaultProps()
        return sn.snapshot_properties(sn=sn, req=req, req_not=req_not,
                                      req_match=req_match, req_match_not=req_match_not)

    @classmethod
    def snapshot_properties_diff(cls, sn1=None, sn2=None,
                                req=None, req_not=False,
                                req_match=None,
                                req_match_not=False):
        """ Generate differences between two properties state
        :sn1: first properties state
                default: after loading
        :sn2: second properties state
                default: current properties state
        :req: only keys which have this string
        :req_not: only keys that don't have this string
        :req_match: only keys which match this regular expression
        :req_match_not: only keys which don't match this regex
                
        :returns: (JavaProperties(only in sn1),
                  JavaProperties(sn1, sn1,sn2 differ),
                  JavaProperties(sn2, sn1,sn2 differ),
                  JavaProperties (only in sn2)) tuple
        """
        if sn1 is None:
            sn1 = cls.loadedProps
        if sn2 is None:
            sn2 = cls.snapshot_properties()
        cls.properties_diff_sn = sn2    
        jp_in_sn1 = JavaProperties()
        jp_sn1_differ = JavaProperties()
        jp_sn2_differ = JavaProperties()
        jp_in_sn2 = JavaProperties()
        for prop_key in sn1.getPropKeys():
            if sn1.is_our_key(prop_key, req=req, req_not=req_not, req_match=req_match, req_match_not=req_match_not):
                if sn2.hasProp(prop_key):
                    sn1_prop_val = sn1.getProperty(prop_key, None)
                    sn2_prop_val = sn2.getProperty(prop_key, None)
                    if sn1_prop_val != sn2_prop_val:
                        jp_sn1_differ.setProperty(prop_key, sn1_prop_val)
                        jp_sn2_differ.setProperty(prop_key, sn2_prop_val)
                else:
                    jp_in_sn1.setProperty(prop_key, sn1.getProperty(prop_key, None))
        for prop_key in sn2.getPropKeys():
            if sn2.is_our_key(prop_key, req=req, req_not=req_not, req_match=req_match, req_match_not=req_match_not):
                if not sn1.hasProp(prop_key):
                    jp_in_sn2.setProperty(prop_key, sn2.getProperty(prop_key, None))
        return (jp_in_sn1, jp_sn1_differ, jp_sn2_differ, jp_in_sn2)
        
               
    @classmethod
    def properties_change_print(cls, sn1=None, sn2=None,
                                 req=None, req_not=False,
                                 req_match=None,
                                 req_match_not=False,
                                 prefix=""):
        """ Print out properties change
        :sn1: first properties state
                default: after loading
        :sn2: second properties state
                default: current properties state
        :req: only keys which have this string
        :req_not: only keys that don't have this string
        :req_match: only keys which match this regular expression
        :req_match_not: only keys which don't match this regex
        :prefix: text before each line print, space added if present
        """
        if prefix != "":
            prefix += " "   # add separation
        (jp_in_sn1, jp_sn1_differ, jp_sn2_differ, jp_in_sn2) = cls.snapshot_properties_diff(sn1=sn1, sn2=sn2,
                                                        req=req, req_not=req_not,
                                                        req_match=req_match,
                                                        req_match_not=req_match_not)
        props_in_sn1 = jp_in_sn1.getPropKeys()
        props_in_sn2 = jp_in_sn2.getPropKeys()
        props_sn1_differ = jp_sn1_differ.getPropKeys()  # Same keys as sn2_differ
        if len(props_in_sn1) == 0 and len(props_in_sn2) == 0 and len(props_sn1_differ) == 0:
            cls.lg(f"{prefix}No differences")
            return
        
        if len(props_in_sn1) > 0:
            SlTrace.lg("Only in first:")
            for prop_key in props_in_sn1:
                val = jp_in_sn1.getProperty(prop_key, None)
                cls.lg(f"{prefix}    {prop_key} = {val}")
        
        if len(props_in_sn2) > 0:
            SlTrace.lg("Only in last:")
            for prop_key in props_in_sn2:
                val = jp_in_sn2.getProperty(prop_key, None)
                cls.lg(f"{prefix}    {prop_key} = {val}")
        
        if len(props_sn1_differ) > 0:
            SlTrace.lg("Changed:")
            for prop_key in props_sn1_differ:
                val1 = jp_sn1_differ.getProperty(prop_key, None)
                val2 = jp_sn2_differ.getProperty(prop_key, None)
                cls.lg(f"{prefix}    {prop_key} : {val1} => {val2}")
            
        
    @classmethod
    def properties_properties_prev_sn(cls):
        """ Get previously properties diff snapshot
        """
        return cls.properties_diff_sn
     
    @classmethod
    def onexit(cls):
        print("onexit")
        pgm_duration = cls.time_since()
        cls.lg(f"Program duration: {pgm_duration:.3f} sec")
        cls.save_propfile()
        try:
            if not cls.closed and cls.logWriter is not None:
                abs_logName = os.path.abspath(cls.logName)
                cls.lg("Closing log file %s"
                    % abs_logName)
                cls.logWriter.close()
                cls.closed = True
                cls.logWriter = None        # Flag as not available
        except IOError as e:
            cls.closed = True
            tbstr = traceback.extract_stack()
            cls.lg("Close logfile %s failed %s - %s"
                    % (abs_logName, tbstr), str(e))
        cls.closed = True
        cls.runningJob = False
        if cls.mw is not None and cls.mw_standalone:
            cls.mw.destroy()

    @classmethod
    def getTs(cls, dp=0):
        """
        Get / generate time stamp Format: YYYYMMDD_HHMMSS
        :dp: seconds decimal places default: 0
        """
        tsfmt = "%Y%m%d_%H%M%S"
        if dp > 0:
            tsfmt = "%Y%m%d_%H%M%S_%f"
        ts = datetime.now().strftime(tsfmt)
        if dp > 0:
            if dp >= 6:
                pass        # Whole string
            else:
                ts = ts[:dp-6]      # remove portion of _[dddddd]
        return ts

    @classmethod
    def ts_to_datetime(cls, ts):
        """ Convert time stamp to datetime object
        :ts: time stamp "%Y%m%d_%H%M%S" or "%Y%m%d_%H%M%S_%f"
        :returns: datetime object
        """
        tsfmt = "%Y%m%d_%H%M%S"
        base_len = 15           # No "_dddddd"
        if len(ts) > base_len:
            ddig = base_len - len(ts) -1        # omit "_" separator
            tsfmt += "_%f"          # Get decimal fraction
            ts += ddig * "0"        # get 6 digits of uSec
        dtobj = datetime.strptime(ts, tsfmt)
        return dtobj
    
    @classmethod
    def ts_diff(cls, ts1=None, ts2=None):
        """ Time, in seconds, between first time (ts1) and second time
        :ts1: first time stamp
        :ts2: second time stamp
        :returns: difference in seconds (float)
        """
        t1 = cls.ts_to_datetime(ts1)    
        t2 = cls.ts_to_datetime(ts2)
        diff = (t2-t1).total_seconds()
        return diff


    @classmethod
    def setProps(cls, propName=None, newExt=None, update=None,
                 logtoScreen=None):
        """
        Set up based on Java style properties file
        :propName: given properties file name
            default: generated from script
        :newExt: Extension of created properties file
            default: Created properties file is same as source propertites
        :update: write out updated properties file a end (save_propfile)
        :logtoScreen: True - write also to screen
                      False - only to file  
                default: do default
        @throws IOException
        """
        if propName is not None:
            cls.propName = propName
        cls.newExt = newExt
        if update is not None:
            cls.update = update
        
        if cls.propName is None:
            script_name = os.path.basename(sys.argv[0])
            script_name = os.path.splitext(script_name)[0]
            cls.propName = script_name
            
        if os.path.splitext(cls.propName)[1] == "":
            cls.propName += ".properties"
            
        if not os.path.isabs(cls.propName):
            cwd = os.getcwd()
            dir_base = os.path.basename(cwd)
            if dir_base == "src":
                prop_dir = os.path.dirname(cwd)
                cls.propName = os.path.join(prop_dir, cls.propName)
            else:
                cls.propName = os.path.abspath(cls.propName)
        loadedProps = cls.loadedProps   # Check if newly loaded    
        cls.defaultProps = JavaProperties(cls.propName)
        cls.loadedProps = cls.defaultProps.copy()      # Save for possible comparison
        cls.loadTraceFlags()        # Populate flags from properties
        if loadedProps is None:
            SlTrace.lg(f"Trace levels from properties file {cls.propName}")
            cls.listFlags(logtoScreen=logtoScreen)


    @classmethod
    def getLogPath(cls):
        """ Returns absolute log file path name
        """
        return os.path.abspath(cls.logName)


    @classmethod
    def getPropKeys(cls, snapshot=None, startswith=None):
        """
        get properties keys
        :snapshot: properties snapshot
                default:current properties
        :startswith: starting with this string
                default: All keys
        """
        if snapshot is None:
            snapshot = cls.getDefaultProps()
        props = snapshot.getPropKeys(startswith=startswith)
        return props

    @classmethod
    def getMemory(cls):
        """ Get current memory usage for this process
        """
        import importlib
        try:
            psutil = importlib.import_module("psutil")
        except:
            cls.lg("getMemory - needs pip install of psutil")
            sys.exit(1)
            
        process = psutil.Process(os.getpid())
        mem = process.memory_info().rss  # in bytes
        cls.mem_used_change = mem - cls.mem_used
        cls.mem_used = mem
        return cls.mem_used 

    @classmethod
    def getMemoryChange(cls):
        """ Get current memory usage change since last
        getMemory call NOTE: no check on memory done here
        """
        return cls.mem_used_change


    @classmethod
    def getPropPath(cls):
        """ Returns absolute Properties file path name
        """
        defaultProps = cls.getDefaultProps()
        abs_propName = defaultProps.get_path()
        return abs_propName


    @classmethod
    def getPropPathSaved(cls):
        return cls.propPathSaved
    
    
    @classmethod
    def getLevel(cls, trace_name):
        trace_name = trace_name.lower()
        cls.recordTraceFlag(trace_name)
        if trace_name not in cls.traceFlags:
            v = 0
            cls.traceFlags[trace_name] = v
        v = cls.traceFlags[trace_name]
        if v is None:
            v = 0            # Not there == 0
        return v



    @classmethod
    def setLevel(cls, trace_name, level=True):
        trace_name = trace_name.lower()
        cls.recordTraceFlag(trace_name, level=level)
        flag_key = cls.getTraceFlagKey(trace_name)
        cls.setProperty(flag_key, level)
        if isinstance(level, str):
            try:
                level = str2bool(level)
            except:
                try:
                    level = str2val(level, 0)
                except:
                    pass            # Leave unchanged
        if trace_name == "all":
            cls.traceAll = level
        else:
            cls.traceFlags[trace_name] = level


    @classmethod
    def traceVerbose(cls, level=1):
        return cls.trace("verbose", level=level)


    @classmethod
    def getTraceFlags(cls):
        """
        return array of set trace flags
        """
        return cls.traceFlags.keys()

    
    
    """
    Record flag for saving
    Note value may change
    """
    @classmethod
    def recordTraceFlag(cls, flag, level=1):
    ###    if flag in cls.recTraceFlags:
    ###       return                # Already here - don't
        
        cls.recTraceFlags[flag] =  level
        flag_key = cls.getTraceFlagKey(flag)
        cls.setProperty(flag_key, level, onlyifnew=True)

    @classmethod
    def report(cls, msg, log_it=True):
        """ report message in dialog box
        :msg: message to report
        :log_it: also print msg to log
                default: log message
        """
        if log_it:
            cls.lg(msg)
        if cls.mw is None:
            cls.mw_standalone_mw = True
            cls.mw = tk.Tk()       # create one
        
        SelectReport(master=cls.mw, message=msg)
                 
    @classmethod
    def getTraceValueFromProp(cls, name):
        """
        Return flag value, -1 if none
        """
        flagstr = cls.getProperty(cls.getTraceFlagKey(name))
        if flagstr == "":
            return -1

        return int(flagstr)


    @classmethod
    def loadTraceFlags(cls):
        """
        load trace flags from properties
        """
        defaultProps = cls.getDefaultProps()
        propfile = defaultProps.get_path()
        props = defaultProps.get_properties()
        pattern = r"^\s*" + cls.traceFlagPrefix + r"\.(.*)"
        r = re.compile(pattern)
        ### TBD I need to think about what is going on here
        keys =  list(props.keys())
        for key in keys:
            m = r.match(key)
            if m:
                post_prefix = m[1]
                name = post_prefix        # stuff after prefix
                value = props[key]  # Default to property value
                mn = re.match(r'(\w+)\s*=\s*(\S.*)', post_prefix)
                if mn:
                    name = mn[1]
                    value = mn[2]
                    break
                if re.match(r'<function.*>$',value):
                    SlTrace.lg(f"Skipping trace flag {key} = {value} because functions need re-setup")
                    continue
                cls.setLevel(name, value)

    
    @classmethod
    def getAllTraceFlags(cls):
        return cls.recTraceFlags.keys()

    @classmethod
    def listTraceFlagValues(cls, flags=None, logtoScreen=None):
        if flags is None:
            flags = cls.getAllTraceFlags()
        for flag in flags:
            level = cls.getLevel(flag)
            cls.lg(f"{flag} = {level}", to_stdout=logtoScreen)

    @classmethod
    def listFlags(cls, logtoScreen=None):
        all_flags = cls.getAllTraceFlags()
        for flag in all_flags:
            level = cls.getLevel(flag)
            cls.lg(f"{flag} = {level}", to_stdout=logtoScreen)

    
    @classmethod
    def getTraceFlagKey(cls, name):
        key = cls.traceFlagPrefix + "." + name
        return key


    @classmethod
    def setDebug(cls, level=1):
        """
        @param debug
                   to set
        """
        cls.setLevel("debug", level)


    @classmethod
    def setVerbose(cls, level=1):
        """
        @param level
        """
        cls.setLevel("verbose", level)


    @classmethod
    def getVerbose(cls):
        """
        @return the verbose
        """
        return cls.getLevel("verbose")


    """
    Trace if at or above this level
    if level is present, flag is of type level
    """
    @classmethod
    def trace(cls, flag=None, level=None, default=None):
        """ trace flag
        :flag: string identifying interest
            default: pass (return True)
        :level: test level, type used as flag type
                boolean: check for ==, None => ck for True
                int, float, str: for >=
                default: boolean True
        :default: if present and flag is new set to this value
        :initialize: set to level immediately
                    default: set False or zero or ""
        :returns: True if passing, level if default
        """
        
        if flag is None:
            return True

        if default is not None and flag in cls.traceFlags:
            default_type = type(default)
            flag_val = cls.getLevel(flag)
            flag_type = type(flag_val)
            if callable(default) and callable(flag_val):
                return              # No change required
            
            if default_type != flag_type:
                cls.lg(f"default({default} type:{default_type}"
                           f" != flag_type: {flag_type} - changing flag")
                cls.setLevel(flag, default)
                return
            
        if flag not in cls.traceFlags:
            if default is not None:
                cls.traceFlags[flag] = default
                return default

            if level is None:
                v = False
            elif type(level) == bool:
                v = False
            else:
                if type(level) == int:
                    v = 0
                elif type(level) == float:
                    v = 0.0
                elif type(level) == str:
                    v = ""
                else:
                    v = level
            cls.traceFlags[flag] = v
        
        fv = cls.traceLevel(flag)
        if type(fv) == bool:
            if level is None:
                level = True
            return  fv == level

        if level is not None: 
            return  fv >= level

        return fv

    @classmethod
    def traceButton(cls, flag, command):
        """ Setup method to be called when button works
        Display and call is done in another module,
        e.g. trace_control_window(TraceControlWindow)
        :flag: identifying flag string
        :command: command function to be called with flag as an argument
        """
        cls.trace(flag, default=command)

    
    """
    Return trace level
    """
    @classmethod
    def traceLevel(cls, flag):
        traceLevel = cls.getLevel(flag)
        return traceLevel


    @classmethod
    def deleteProperty(cls, key):
        """ Delete property from properties file
        """
        defaultProps = cls.getDefaultProps()
        if cls.hasProp(key):
            defaultProps.deleteProperty(key)


    @classmethod
    def getProperty(cls, key, default=""):
        """    
        :key:property key
        :default: return if Not found default: ""
        ###@return property value, "" if none
        """
        defaultProps = cls.getDefaultProps()
        return defaultProps.getProperty(key, default)


    @classmethod
    def setProperty(cls, key, value, onlyifnew=False):
        defaultProps = cls.getDefaultProps()
        if not cls.hasProp(key) or not onlyifnew:
            defaultProps.setProperty(key, value)

    @classmethod
    def getLogFile(cls):
        """ Returns output file to support situations
        which can't be logged directly via lg
        Note that no time stamps will be shown
        """
        return cls.setupLogging()

    @classmethod
    def getSourceDirs(cls, string=False):
        """ Returns list of source directories
        """
        dirs = cls.getAsStringArray("source_files")
        dirpaths = []
        for dir in dirs:
            dirpaths.append(os.path.abspath(dir)) 
        if string:
            return " ".join(dirpaths)
        
        return dirpaths
    
    @classmethod
    def getSourcePath(cls, fileName, req=True, report=True):
        """
        Get source absolute path Get absolute if fileName is not absolute TBD handle
        chain of paths like C include paths
        
        :fileName:
        :req: Required default: Must be found or error or raise TraceError
        :report: Report if can't find default: report to log,stderr if can't find
        :return: absolute file path if found, else None
        """
        if not os.path.isabs(fileName):
            dirs = cls.getAsStringArray("source_files")
            searched = []
            for pdir in dirs:
                inf = os.path.join(pdir, fileName)
                inf = os.path.abspath(inf)
                if os.path.exists(inf):
                    try:
                        return os.path.realpath(inf)
                    except IOError as e:
                        raise TraceError("Problem with path %s" % (inf))

                searched.append(os.path.realpath(inf))
            if report:
                cls.lg("%s was not found" % fileName)
                if len(dirs) > 0:
                    cls.lg("Searched in:")
                    for dirp in dirs:
                        dirpath = os.path.abspath(dirp)
                        cls.lg("\t%s" % dirpath)
            if req:
                raise TraceError("Can't find file %s in %s" % (fileName, dirs))
            
            return None
        # Return unchanged
        if not os.path.exists(fileName):
            if report:
                cls.lg("Can't find file %s" % (fileName))
            if req:    
                raise TraceError("Can't find file %s" % (fileName))
            
            return None
            
        return fileName # Already absolute path

    @classmethod
    def getIncludePath(cls, fileName):
        """
        Get include absolute path Get absolute if fileName is not absolute TBD handle
        chain of paths like C include paths
        
        @param fileName
        @return absolute file path, "" if not found
        """
        if not os.path.isabs(fileName):
            dirs = cls.getAsStringArray("include_files")
            searched = []
            for dirp in dirs:
                inf = os.path.join(dir, fileName)
                if os.path.exists(inf) and not os.path.isdir(inf):
                    try:
                        return os.path.realpath(inf)
                    except IOError as e:
                        print("Problem with path %s,%s" % (dirp, fileName), file=sys.stderr)
      
            print("%s was not found\n" % fileName, file=sys.stderr)
            if len(dirs) > 0:
                print("Searched in:\n")
                for dirp in dirs:
                    dirpath = os.path.realpath(dirp)
                    print("\t%s\n" % dirpath)
            return "" # Indicate as not found

        return fileName # Already absolute path


    @classmethod
    def getAsStringArray(cls, propKey):
        """
        Get default properties key with value stored as comma-separated values as an
        array of those values If propKey not found, return an empty array
        
        @param propKey
        @return array of string values
        """
        valstr = cls.getProperty(propKey)
        vals = valstr.split(',')
        return vals


    @classmethod
    def setTraceFlag(cls, trace_name, level=1):
        """
        Fast, lowlevel trace level setting
        Used for interactive trace changes
        @param trace_name
        @param level
        """
        cls.traceFlags[trace_name] = level

    @classmethod
    def hasProp(cls, key):
        """ Check if property is already present
        :key: property key
        :returns: True iff property is present
        """
        defaultProps = cls.getDefaultProps()
        return defaultProps.hasProp(key)
            
    
if __name__ == "__main__":
    import argparse
    propfile = "none_expected"
    propfile_new_ext = "prop_out"
    propfile_update = True
    show_passes = True
    ###show_passes = False
    quit_on_fail = True      # True - stop testing on first failure
    
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--propfile', default=propfile,
                        help=("properties file name"
                              " (default:generate from script name)"))
    parser.add_argument('-n', '--propfile_new_ext', default=propfile_new_ext,
                        help=("new properties file ext"
                              " (default:no extension change)"))
    parser.add_argument('-u', '--propfile_update', default=propfile_update,
                        help=("update properties file"
                              " (default:write out new properties file)"))
    parser.add_argument('-s', '--show_passes', action='store_true', default=show_passes,
                        help=(f"Show passes"
                              f" (default: {show_passes}"))
    parser.add_argument('-q', '--quit_on_fail', action='store_true', default=quit_on_fail,
                        help=(f"Quit on first failure"
                              f" (default: {quit_on_fail}"))

    args = parser.parse_args()             # or die "Illegal options"
    
    propfile = args.propfile
    propfile_new_ext = args.propfile_new_ext
    propfile_update = args.propfile_update
    show_passes = args.show_passes
    quit_on_fail = args.quit_on_fail
        
    def test_fail(msg):
        """ Show error and quit if quit_on_first_error
        """
        SlTrace.lg("FAIL: %s" % msg)
        if quit_on_fail:
            SlTrace.lg("Quitting on first fail")
            sys.exit(1)

            
    def test_pass(msg):
        """ Show pass if show_passes
        """
        print_msg = "Pass: %s" % msg
        if show_passes:
            print(print_msg)
        SlTrace.lg(print_msg, to_stdout=False)
            
    logName = "sltest"
    SlTrace.setupLogging(logName, propName=propfile)     # Setup log/properties names
    SlTrace.setProps(newExt=propfile_new_ext, update=propfile_update)
    SlTrace.lg("args: {}\n".format(args))
    SlTrace.lg("setupTest()")
    flag = "sf1"
    SlTrace.setLevel(flag)
    level_got = SlTrace.getLevel(flag)
    if level_got != 1:
        print("level_got: %d != 1" % level_got)
        sys.exit(1)
    flag = "sf2"
    level = 3
    SlTrace.setLevel(flag, level)
    level_got = SlTrace.getLevel(flag)
    if level_got != level:
        test_fail("level_got: %d != %d" % (level_got, level))

    
    if SlTrace.trace(flag, level):
        test_pass("traced %s(%s) with trace(%d)" % (flag, level_got, level))
    else:
        test_fail("FAILED trace %s(%s) with trace(%d)" % (flag, level_got, level))
    
    if SlTrace.trace(flag, level_got-1):
        test_pass("Correctly traced %s(%s) with trace(%d)" % (flag, level_got, level_got-1))
    else:
        test_fail("ERRONEOUSLY skipped trace %s(%s) with trace(%d)" % (flag, level_got, level_got-1))
    
    if SlTrace.trace(flag, level_got+1):
        test_fail("ERRONEOUSLY traced %s(%s) with trace(%d)" % (flag, level_got, level_got+1))
    else:
        test_pass("Correctly skipped trace %s(%s) with trace(%d)" % (flag, level_got, level_got+1))
        
    SlTrace.setFlags("f1=1,f2=2,f3=3,fnone")
    tdict = {"f1" : 1, "f2" : 2, "f3" : 3, "fnone" : None}
    for flag, level in tdict.items():
        if level is None:
            level = 1           # Default setting
        trace_level = level
        if SlTrace.trace(flag, trace_level):
            if trace_level is None:
                test_pass("flag %s(%d) traced(None)" % (flag, level))
            else:
                test_pass("flag %s(%d) traced(%d)" % (flag, level, trace_level))
                
        else:
            if trace_level is None:
                test_fail("flag %s(%d) skipped(None)" % (flag, level))
            else:
                test_fail("flag %s(%d) skipped(%d)" % (flag, level, trace_level))
    SlTrace.lg("Making the following output only to log")
    SlTrace.setupLogging(logToScreen=False, dp=6)
    for i in range(250):
        SlTrace.lg(f"{i}: only to log file")
    SlTrace.setupLogging(logToScreen=True, dp=0)
    SlTrace.lg("Output is now to both log file and screen")
    SlTrace.setupLogging(stdOutHasTs=True)
    SlTrace.lg("Timestamp now included on screen")
    SlTrace.setupLogging(dp=4)
    SlTrace.lg("with new timestamp precision")
    SlTrace.lg("More of the same")
    SlTrace.setupLogging(dp=0)
    SlTrace.lg("dp back at 0")
    SlTrace.setLogStdTs(False)
    SlTrace.lg("No timestamp to STDOUT")
    test_memory = False
    test_memory = True
    if test_memory:
        mem = SlTrace.getMemory()
        SlTrace.lg(f"memory = {mem} bytes")
    SlTrace.lg("End of Test")



# grid_fill_gobble.py    23Nov2022, crs
"""
Grid Fill Gobble Algorithm
First Algorithm: "gobble - keep adding 
all adjacent ungroupped cells to to check_group,
till no ungrouped adjacent cells can be added.

"""
from grid_rot import GridRot as GR

class GridFillGobble:

    def __init__(self, cells, excluded=None):
        """ Setup gobble strategy
        :cells: dictionary (only the keys are used here)
                of occupied cells
        :excluded: extra cell exclusion list
                    dictionary, only the keys are used
        """
        self.cells = cells
        if excluded is None:
            excluded = []
        self.excluded = excluded
        self.travel_list = []   # List of cells (ixy),
                                # traversal order
                                # reset in find_region
        
    def find_region(self, ixy):
        """ Find a list of occupied cells,
        from self.cells which start with ixy,
        which cover a connected region
        Hopefully we get a good traversal list.

        :ixy: (ix,iy) tuple for one enclosed cell
        :returns: list of (ix,iy) in traversal order
        """
        self.checking_cells = {} # of ixy
                            # where:
                            # ixy - cell xy
        self.travel_list = [] # List in traversal order
        self.add_ck_cell(ixy)
        while self.check_if_filled():
            pass
        return self.travel_list

    def check_if_filled(self):
        n_added = 0      # Sum added
        for cell in list(self.checking_cells):
            neighbors_added = self.check_cell_neighbors(
                                     cell)
            n_added += neighbors_added
        if n_added > 0:
            return True 
        
        return False

    def check_cell_neighbors(self, cell):
        """ Check cell's neighbors
        for possible adding to region
        For each neighbor, if appropriate
            Add it
        :cell: who's neighbors to check
        :returns: number of cells added
        """
        n_added = 0
        cell_ixy = cell
        for idir in range(len(GR.idirs)):
            cell_n = GR.next_cell(cell_ixy, idir)
            if self.is_cell_part(cell_n):
                self.add_ck_cell(cell_n)
                n_added += 1
        return n_added
        
    def is_cell_part(self, ixy):
        """ Check if location is suitable for new addition
        to region.
        :checking_cells: dictionary of cells checked or
                in line to be checked
        :ixy: cell ixy to be checked
        :excluded: cells to not be considered
        :returns: True iff to be added else False
        """
        if ixy in self.checking_cells:
            return False# Already covered
        
        if ixy in self.excluded:
            return False # excluded
        
        if ixy not in self.cells:
            return False # Part of figure
        
        return True

    def add_ck_cell(self, ixy):
        """ Add new cell to fill region
        Currently just ixy.  At sometime we might include
        additional information
        
        :ixy:  cell to be added
        """
        self.checking_cells[ixy] = ixy
        self.travel_list.append(ixy)

if __name__ == "__main__":
    print(f"Self Test: {__file__}")
    cells_set = {
            (2,6), (3,6), (4,6), (5,6), (6,6),
            (2,5),               (5,5),
            (2,4),               (5,4),
            (2,3), (3,3), (4,3), (5,3),
            (2,2), (3,2), (4,2), (5,2), (6,2),
         }
    cells = dict.fromkeys(cells_set, 0)
    start_cell = (2,2)
    grfill = GridFillGobble(cells=cells, excluded={(4,3)})
    travel_cells = grfill.find_region(start_cell)
    print("travel_cells:", travel_cells)
    travel_cells_order = {}  # by ixy order number
    order = 0
    for tc in travel_cells:
        order += 1
        travel_cells_order[tc] = order
    for irow in reversed(range(10)):
        for icol in range(10):
            tc = (icol,irow)
            if tc in travel_cells_order:
                order = travel_cells_order[tc]
                print(f"({icol},{irow})/{order:<2} ", end="")
            else:
                print(" "*14, end="")
        print()



#wx_stuff.py    04Apr2024
""" support functions
"""
import wx

def wx_Point(x, y=None):
    """ Force int args
    :x: x value or wx.Point if y is None
    :y: y value
    :returns: wx.Point
    """
    if y is None:
        return x
    
    return wx.Point(int(x), int(y))



#wx_trace_control_panel.py  24Apr2024  crs, move ctl buttons
""" 
Control buttons / flags for trace flag control window
"""
import wx
import wx.lib.scrolledpanel as scrolled

from select_trace import SlTrace
from wx_trace_flag_box import TraceFlagBox


class FlagsPanel(scrolled.ScrolledPanel):

    def __init__(self, parent):

        scrolled.ScrolledPanel.__init__(self, parent, -1)

        vbox = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)
        self.vbox = vbox
                
        ###self.flags_panel = wx.Panel(self)
        ###self.flags_sizer = wx.BoxSizer(wx.VERTICAL)
        ###self.flags_panel.SetSizer(self.flags_sizer)
        
    def add(self, entry):
        """ Add flag  entry
        :entry: (TraceFlagBox)
        """
        self.vbox.Add(entry, 0)
        
    def clear(self):
        """ clear out entries
        """
        self.vbox.Clear(delete_windows=True)
    
    def complete(self):
        """ Complete scrolling setup
        """
        self.SetupScrolling(scroll_x=False)
        #self.Fit()


class TraceControlPanel(wx.Panel):
    def __init__(self, panel, change_call=None):
        """ Support trace control display and updates
        :change_call: function to do updates default: no calls
        """
        self.flags_sizer = None     # sizer controling flags
        self.flag_by_cb = {}             # Dictionary hashed on cb widget
        self.data_by_flag = {}      # [tcb, flag]
        
        self.standalone = False      # Set True if standalone operation
        self.change_call = change_call
        
        wx.Panel.__init__(self, panel)
        
        ctl_sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.v_sizer = wx.BoxSizer(wx.VERTICAL)

        tc_all_button = wx.Button(self, label="SET ALL")
        tc_all_button.Bind(wx.EVT_BUTTON, self.set_all)
        tc_clear_button = wx.Button(self, label="CLEAR ALL")
        tc_clear_button.Bind(wx.EVT_BUTTON, self.clear_all)
        tc_bpt_button = wx.Button(self, label="BPT")
        tc_bpt_button.Bind(wx.EVT_BUTTON, self.breakpoint)
        tc_show_all_button = wx.Button(self, label="Show ALL")
        tc_show_all_button.Bind(wx.EVT_BUTTON, self.show_all)
        tc_show_set_button = wx.Button(self, label="Show SET")
        tc_show_set_button.Bind(wx.EVT_BUTTON, self.show_set)
        tc_show_clear_button = wx.Button(self, label="Show CLEAR")
        tc_show_clear_button.Bind(wx.EVT_BUTTON, self.show_clear)

        self.flags_panel = FlagsPanel(self)  # Holds flags panel
        
        ctl_sizer.Add(tc_all_button, 1, wx.RIGHT)        
        ctl_sizer.Add(tc_clear_button, 1, wx.RIGHT)
        ctl_sizer.Add(tc_bpt_button, 1, wx.RIGHT)
        ctl_sizer.Add(tc_show_all_button, 1, wx.RIGHT)
        ctl_sizer.Add(tc_show_set_button, 1, wx.RIGHT)
        ctl_sizer.Add(tc_show_clear_button, 1, wx.RIGHT)
        self.v_sizer.Add(ctl_sizer, 0, wx.EXPAND)
        self.v_sizer.Add(self.flags_panel, 1, wx.EXPAND)
        self.SetSizer(self.v_sizer)

        flags = self.getAllFlags()
        self.display_flags(flags)
        self.v_sizer.Fit(self)
        self.Show()

    def list_ckbuttons(self):
        cb_flags = sorted(self.data_by_flag.keys())
        for flag in cb_flags:
            tcb = self.data_by_flag[flag][0]
            value = tcb.get() if tcb is not None else None
            SlTrace.lg(f"flag={flag} val={value}")

    def display_flags(self, flags):
        """Display new set of flags
        :flags: ordered list of flag strings
        """
        self.data_by_flag = {}  # Set to displayed controls
        self.flags_panel.clear()
        for flag in flags:
            value = self.getLevel(flag)
            tfb = TraceFlagBox(self.flags_panel, flag, value=value)
            self.data_by_flag[flag] = [tfb, flag]
            self.flags_panel.add(tfb)
        self.flags_panel.complete()
        self.Layout()    
        self.Show()
        
    def show_all(self, event=None):
        """ Select buttons to show
        """
        flags = sorted(self.getAllFlags())
        self.display_flags(flags)

    def show_set(self, event=None):
        """ Select buttons to show
        """
        flags = sorted(self.getAllFlags())
        set_flags = []
        for flag in flags:
            if self.getLevel(flag):
                set_flags.append(flag)
        self.display_flags(set_flags)

    def show_clear(self, event=None):
        """ Select buttons to show
        """
        flags = sorted(self.getAllFlags())
        clear_flags = []
        for flag in flags:
            if not self.getLevel(flag):
                clear_flags.append(flag)
        self.display_flags(clear_flags)
        
    def set_all(self, event=None):
        """ All all trace flags
        """
        for flag in sorted(self.getAllFlags()):   # In display order
            self.set_trace_level(flag, True)
        self.show_set()

    def clear_all(self, event=None):
        """ Clear all known trace flags (only booleans)
        """
        for flag in sorted(self.getAllFlags()):   # In display order
            self.set_trace_level(flag, False)
        self.show_clear()
        
    def set_trace_level(self, flag, val, change_cb=True):
        """ Set trace level, changing Control button if requested
        :flag: - trace flag name
        :val: - value to set
        :change_cb: True(default) appropriately change the control
        """
        self.setLevel(flag, val)
        
        if flag not in self.data_by_flag:
            return
        
        tcb, flag = self.data_by_flag[flag]        
        if tcb is None:
            SlTrace.lg("set_trace_level(%s,%d) - flag No tcb" % (flag, val))
            return

        if change_cb:
            tcb.set(flag, val)
                    
        SlTrace.lg(f"flag={flag}, val={val}", "trace_flags")
        self.setLevel(flag, val)
            
        if self.change_call is not None:
            self.change_call(flag, val)

    def getAllFlags(self):
        return SlTrace.getTraceFlags()
    
    def getLevel(self, flag):
        return SlTrace.getLevel(flag)

    def setLevel(self, flag, value):
        """ Set level
        :flag: flag string
        :value: value
        :returns: new value
        """
        SlTrace.setLevel(flag, value)
        return value
    
    def breakpoint(self, event=None):
        """ Force immediate breakpoint - enter debugger
        """
        import pdb
        SlTrace.lg("Breakpoint")
        pdb.set_trace()
            
########################################################################
class MainFrame(wx.Frame):
    """"""

    #----------------------------------------------------------------------
    def __init__(self):
        """Constructor"""
        wx.Frame.__init__(self, None,
                          title="Testing TraceControlPan",
                          size=(600,400))
        panel = TraceControlPanel(self)
        self.Show()

#----------------------------------------------------------------------
if __name__ == "__main__":
    SlTrace.lg("Testing")
    nflags = 50
    for n in range(1,nflags+1):
        flag = f"flag_{n}"
        SlTrace.setTraceFlag(flag, True)
    app = wx.App(False)
    frame = MainFrame()
    app.MainLoop()

 


# x_show_square_loop_colors_braille.py  # braille
# Display a square with colored sides
#isolate except for this directory
import sys
import os

print(f"Initial sys.path:\n    {"\n    ".join(sys.path)}")
src_file = os.path.abspath(__file__)
print(f"add_src_dir file: {src_file}")
src_path = os.path.dirname(src_file)
if src_path not in sys.path:
    print(f"Inserting source directory {src_path}")
    sys.path.insert(0, src_path)
    print(f"sys.path:\n    {"\n    ".join(sys.path)}")

file_dir = os.path.dirname(__file__)
def remove_path_ending(ending=None):
    """ Remove all sys.path entries ending in 
    :ending: ending of path to be removed
    """
    found = True
    while found:
        found = False       # Set if any found
        sys_path = sys.path
        for ent in sys.path:
            if ent.endswith(ending):
                sys_path.remove(ent)
                print(f"Removing: {ent}")
                found = True
        if found:
            sys.path = sys_path
                
remove_path_ending(r"\resource_lib\src")
if len(sys.argv) > 1:
    hack_dir = r"c:\Users\raysm\vscode\testing\venv\Lib\site-packages\graphics_braille"
    print(f"adding: {hack_dir}")
    sys.path.append(hack_dir)
print(f"sys.path:\n    {"\n    ".join(sys.path)}")
##from turtle import *		     # Get standard stuff
from wx_turtle_braille import *		     # Get braille

colors = ["red","orange","yellow","green"]

for colr in colors:
    width(40)
    color(colr)
    forward(200)
    right(90)
done()		    # Complete drawings



# canvas_copy.py    22Jan2025  crs
"""
Approximate deep copy of tkinter Canvas object
"""
import tkinter as tk

def deep_copy_canvas(canvas, root=None, trace=False):
    """ copy canvas
    :canvas: tk.Canvas
    :root: tk root
            default: create tk.Tk()
    :trace: debug tracing
        default: False
    """
    if root is None:
        root = tk.Tk()
        root.withdraw() # hide window of copy
        
    new_canvas = tk.Canvas(root, width=canvas.winfo_width(), height=canvas.winfo_height())
    new_canvas.pack()
    new_canvas.unsupported = []
    for item_id in canvas.find_all():
        item_type = canvas.type(item_id)
        coords = canvas.coords(item_id)
        options = canvas.itemconfig(item_id)
        config = canvas.itemconfig(item_id)
        new_config = {key: config[key][-1] for key in config.keys()}
        if trace: print(f"\n{item_type}: coords:{coords}")
        if trace: print(f"{item_type}: options:{options}")
        item = new_canvas._create(item_type, coords, {})
        new_canvas.itemconfigure(item, new_config)

    return new_canvas

def canvas_exclude_types(cv, exclude):
    """ Exclude item types
    :cv: canvas
    :items: list of item ids to use
    :exclude: list of types
            default: ['image']
    :returns: list of non-excluded types
    """
    if exclude is None:
        exclude = ['image']
    ok_items = []
    for item in cv.find_all():
        item_type = cv.type(item)
        if item_type not in exclude:
            ok_items.append(item)
    return ok_items

def is_set_opt(option_list):
    """ Check if opt is set, i.e. not default
    :option: option val list (opt_name, v, v,...)
    """
    is_set = False  # Set if any not ''
    opt_name = option_list[0]
    opt_vals = option_list[1:]
    for opt_val in opt_vals:
        if opt_val != '':
            is_set = True
            break
    return is_set

option_settings_cache = {}  # by type by option        
def canvas_show_items(cv, exclude_types=None, show_coords=True,
                      show_options=False,
                      use_value_cache=True):
    """ Show canvas items
    :exclude_types: list of types to exclude
            default: that of canvas_exclude_types
    :show_coords: show coordinates for each item
    :show_options: show item's options as best we can
    :use_value_cache: True - only list settings if
                    different from option_settings_cache
                    update cache with new settings
    :returns: items in canvas
    """
    res = ""
    items = canvas_exclude_types(cv, exclude_types)
    for item in items:
        item_type = cv.type(item)
        if item_type not in option_settings_cache:
            option_settings_cache[item_type] = {}
        item_option_settings_cache = option_settings_cache[item_type]
        coords = cv.coords(item)
        options = cv.itemconfig(item)
        res += "\n   "  # Each item begins new line indented
        res += f" {item_type}:"
        if show_coords:
            res += f"({coords})"
        if show_options:
            for option in options:
                opt_val_is_new = False
                option_list = options[option]
                opt_name = option_list[0]
                val_list = option_list[1:]   # [0] is option name
                if is_set_opt(option_list):
                    if opt_name not in item_option_settings_cache:
                        opt_val_is_new = True
                        item_option_settings_cache[opt_name] = val_list
                    if (not use_value_cache
                        or opt_val_is_new
                        or val_list != item_option_settings_cache[opt_name]):
                        res += f" {opt_name}={val_list}"
                item_option_settings_cache[opt_name] = val_list  # force updates
            
    return res

def canvas_diff(cv1, cv2, show_same=True, exclude_types=None,
                show_items=False, show_coords=False,
                show_options=False,
                use_value_cache=True):
    """ Find differences between canvases
    :cv1: first canvas
    :cv2: second canvas
    :show_same: Show if same
                default: True
    :show_items: List items
                defalut: False
    :show_coords: show coordinates
                default: False
    :show_options: show items' options
                default: False
    :exclude_types: list of types to exclude
            default: canvas_exclude_types default
    :use_value_cache: True - use value cache to 
                    minimize output
                default: True
    :returns: string with description
    """
    res = ""    # Result string
    items1 = canvas_exclude_types(cv1, exclude_types)
    items2 = canvas_exclude_types(cv2, exclude_types)
    cmp_len = len(items1) - len(items2) # < 0, == 0, > 0
    if cmp_len == 0:
        if show_same:
            res += f"same len: {len(items1)}"
    else:
        res += f"lengths differ"
        res += f" (1): {len(items1)}"
        res += f" (2): {len(items2)}"
    
    if cmp_len == 0:
        nsame = 0
        first_same = -1
        last_same = -1
        first_diff = -1
        first_diff_type1 = ""
        first_diff_coords1 = ""
        first_diff_type2 = ""
        first_diff_coords2 = ""
        last_diff = -1
        for i in range(len(items1)):
            item1_type = cv1.type(i)
            item1_coords = cv1.coords(i)
            item1_options = cv1.itemconfig(i)
            item2_type = cv2.type(i)
            item2_coords = cv2.coords(i)
            item2_options = cv2.itemconfig(i)
            same_type = False
            same_coords = False
            if item1_type == item2_type:
                same_type = True
                if item1_coords == item2_coords:
                    same_coords = True
            if same_type and same_coords:
                nsame += 1
                if first_same == -1:
                    first_same = i
                last_same = i
            else:
                if first_diff < 0:
                    first_diff = i
                    first_diff_type1 = item1_type
                    first_diff_type2 = item2_type
                    first_diff_coords1 = item1_coords
                    first_diff_coords2 = item2_coords
                last_diff = i
        if first_diff < 0:
            if show_same:
                res += " all same type, coords"
        else:
            res += f" first diff at {i}"
            if first_diff_type1 == first_diff_type2:            
                res += f" types both: {first_diff_type1}"
                res += f" coords: (1){first_diff_coords1} (2){first_diff_coords2}"
            else:
                res += " type (1) {first_diff_type1} != (2) {first_diff_type2}"
    if show_items:
        cv1_show = canvas_show_items(cv1, show_coords=show_coords,
                                     show_options=show_options,
                                     use_value_cache=use_value_cache)
        res += f"\n    (1) {cv1_show}"
        cv2_show = canvas_show_items(cv2, show_coords=show_coords,
                                     show_options=show_options)
        res += f"\n    (2) {cv2_show}"                  
    return res

if __name__ == '__main__':
    from turtle import *    # Bring in turtle graphic functions
    
    show_coords = True
    use_cache = True
    #show_coords = False
    #use_cache = False
    
    trace = True
    trace = False
    
    speed("fastest")
    colors = ["red","orange","yellow",
            "green"]

    canvas_copies = []     # snapshots 
    screen = getscreen()
    cv = screen.getcanvas()

    cvc = deep_copy_canvas(cv)
    canvas_copies.append(cvc)
    for colr in colors:
        width(40)
        color(colr)
        forward(200)
        right(90)
        cvc = deep_copy_canvas(cv, trace=trace)
        canvas_copies.append(cvc)
    #done()
    
    #Look for unsupported types in last copy
    for ut in canvas_copies[-1].unsupported:
        print(f"Unsupported type: {ut}")
    
    for i, cvc in enumerate(canvas_copies):
        cv = cvc
        if i < 1:
            cv2 = cv
            continue
        else:
            cv1 = cv2
            cv2 = cvc
        res = canvas_diff(cv1, cv2, show_items=True,
                          show_coords=show_coords,
                          show_options=True,
                          use_value_cache=use_cache)
        print(f"\ncv(diff {i-1} to {i}: {res}")    



#wx_tk_bg_call.py 20Sep2024  crs
""" 
Background caller to facilitate calling functions from within tkinter's main thread
Tkinter requires that some functions be executed in the main thread
"""
import queue
import tkinter as tk
import copy

from select_trace import SlTrace

""" call entry """
class BgCallEntry:
    def __init__(self, bg_call, function, args, kwargs,
                 root=None, call_num=None, ret=None):
        """ Setup call entry
        :bg_call: reference to caller (TgBgCall)
        :function: function to be called from main thread
        :args: pocition args
        :kwargs: key word args
        :call_num: unique call number default: generated
        :ret: function return
        """
        self.bg_call = bg_call
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.root = root
        call_num = bg_call.next_call_num(call_num)
        self.call_num = call_num
        self.ret = ret
        self.done = False       # Set when call complete

    def get_name(self):
        """ Get function name
        """
        return self.function.__name__
       
    def __str__(self):
        st = f"BgCallEntry[{self.call_num}]: {self.get_name()} {self.args} {self.kwargs}"
        return st
        
class TkBgCall:
    
    def __init__(self, root, canvas_grid=None,
                 canvas_grid_snapshots = [], check_queue_int=10):
        """ Setup for background call in which runctions are called
        :root: root (root) object
        :canvas_grid: base canvas
        :check_int: queue check interval (msec)
        """
        self.root = root
        self.canvas_grid = canvas_grid
        self.canvas_grid_snapshots = canvas_grid_snapshots
        self.check_queue_int = check_queue_int
        self.call_num = 0               # Unique call #
        self.call_entry_d = {}          # dictionary of pending calls
        self.call_rets = {}             # Dictionary by call_num of pending call returns
        self.call_queue = queue.Queue()
        self.call_ret_queue = queue.Queue()
        cell_specs = self.canvas_grid.get_cell_specs()
        SlTrace.lg(f"TkBgCall: cell_specs: {cell_specs}", "cell_specs")
        self.set_check_queue()

    def set_check_queue(self):
        """Set check interval"
        """
        self.root.after(self.check_queue_int, self.check_queue)
                
    def check_queue(self):
        """ check for and process any pending commands for background
        """
        SlTrace.lg("check_queue", "check_queue")
        if not self.call_queue.empty():
            call_entry = self.call_queue.get()
            SlTrace.lg(f"check_queue call_entry:{call_entry}", "check_queue")
            self.bg_caller(call_entry)
        self.set_check_queue()    
            

    def next_call_num(self, call_num=None):
        """ Get/adjust next call number
        :call_num: next call number
            default: increase number by one
        :returns: updated call number
        """
        if call_num is None:
            call_num = self.call_num + 1
        self.call_num = call_num
        return self.call_num


    
    def bg_caller(self, call_entry):
        """ Calls function.  Signals when call is complete
        MUST BE CALLED FROM MAIN THREAD
        :call_num: index in self.call_entry_d of function entry 
        """
        SlTrace.lg(f"bg_caller calling {call_entry}", "bg_calling")
        name = call_entry.get_name()
        if name == "get_cell_specs":
            cell_specs = self.canvas_grid.get_cell_specs()
            SlTrace.lg(f"\nTkBgCall:bg_caller cell_specs: {cell_specs}",
                       "cell_specs")
            if "snapshot_num" in call_entry.kwargs:
                snapshot_num = call_entry.kwargs["snapshot_num"]
                SlTrace.lg(f"bg_caller snapshot[{snapshot_num}]", "bg_caller")                
                del call_entry.kwargs["snapshot_num"]
                if snapshot_num is not None and snapshot_num > 0:
                    canvas_grid = self.canvas_grid_snapshots[snapshot_num-1]
                else:
                    canvas_grid = self.canvas_grid
            else:
                canvas_grid = self.canvas_grid
            retorig = canvas_grid.get_cell_specs(*call_entry.args,
                                                      **call_entry.kwargs)
        elif name == "get_rect_tur":
            rect_tur = self.canvas_grid.get_cell_rect_tur()
            SlTrace.lg(f"\nTkBgCall:bg_caller rec_tur: {rect_tur}", "bg_caller")
            retorig = rect_tur
        else:
            retorig = call_entry.function(*call_entry.args, **call_entry.kwargs)
            SlTrace.lg(f"""\n TkBgCall::bg_caller: call_entry: {call_entry}
                        \n retorig: {retorig}
                    """, "bg_caller")
            
        SlTrace.lg(f"TkBgCall.bg_caller: retorig:{retorig}", "bg_caller")
        ret = copy.copy(retorig)
        SlTrace.lg(f"TkBgCall:bg_caller ret: {ret}", "bg_caller")
        self.call_rets[call_entry.call_num] = ret
        call_entry.ret = ret     # Save function return for delayed return
        call_entry.done = True   # Set call completion indicator
        self.call_ret_queue.put(call_entry)
    
    def set_call(self, function, *args, **kwargs):
        """ Setup call function and args in dictionary to be called in the future in bacground
            Do background till function completes
            Return function's return value
        :function: function to call
        :args: positional args
        :kwargs: keyword args
        :returns: unique call nunber
        """
        call_entry = BgCallEntry(self, function, args, kwargs, root=self.root)
        SlTrace.lg(f"call_queue.put {call_entry}", "bg_caller")
        self.call_queue.put(call_entry)
        return call_entry.call_num
       
    def get_ret(self, call_num):
        """ Get call return
        :call_num: unique call number
        :returns: call's return, None if not ready
        """
        if call_num in self.call_rets:
            return self.call_rets[call_num]
        
        return None
    def wait_completion(self, call_num):
        """ Wait till call completed, giving background time to continue
        :call_num: call number
        """
        while True:
            self.root.update()
            call_ret = self.get_ret(call_num)
            if call_ret is not None:
                break

'''                    
    def call(self, *args, **kwargs):
        """ TBD catch unexpected call
        """
        SlTrace.lg(f"Unexpected call {args}, {kwargs}")
'''
        
if __name__ == '__main__':
    def test_fn(root, *args, **kwargs):
        """ Test function
        :root: window
        """
        btn = tk.Button(root, text = f'Click me !\n {args}', 
                command = root.destroy) 

        # Set the position of button on the top of window 
        btn.pack(side = 'top')     
        print(f"test_fn({args}, kwargs={kwargs})")
        return f"returns " f"test_fn({args}, kwargs={kwargs})"

    root = tk.Tk()
            
    bg = TkBgCall(root)
    
    call_num = bg.set_call(test_fn, root,  "a1_val", kwa="kwa_val")
    print(f"call_num={call_num}")
    bg.wait_completion(call_num)
    ret = bg.get_ret(call_num)
    print(f"call_ret: {ret}\n")

    call_num = bg.set_call(test_fn, root, "a1_second_call")
    print(f"call_num={call_num}")
    bg.wait_completion(call_num)
    ret = bg.get_ret(call_num)
    print(f"call_ret: {ret}\n")



# magnify_info.py     18Feb2023   crs, Split from audio_draw_window.py

"""
Information supporting the magnification of a segment of CanvasGrid (canvas_grid.py)
Communication packet between CanvasGrid and AudioDrawWindow
"""
import copy

class MagnifyDisplayRegion:
    """ Region of display, from which a selection may be created """
    def __init__(self, win_fract=True,
                    x_min=None,y_min=None,
                    x_max=None,y_max=None,
                    ncols=40, nrows=25,
                    dup=None):
        """ Setup region of display
        :win_fract: True - values are fraction (0. to 1.) of whole display area
        :x_min: minimum x value - left side 
        :y_min: minimum y value - top side
        :x_max: maximum x value - right side
        :y_max: maximum  y value - bottom side
        :ncols: number of grid columns
        :nrows: number of grid rows
        :dup: if not none use this (MagnifyDisplayRegion) for None entries
        """
        if dup is not None:
            win_fract = dup.win_fract
            if x_min is None:
                x_min = dup.x_min
            if y_min is None:
                y_min = dup.y_min
            if x_max is None:
                x_max = dup.y_max
            if y_max is None:
                y_max = dup.y_max
            if ncols is None:
                ncols = dup.ncols
            if nrows is None:
                nrows = dup.nrows
        self.win_fract = win_fract
        self.x_min = x_min
        self.y_min = y_min
        self.x_max = x_max
        self.y_max = y_max
        self.ncols = ncols
        self.nrows = nrows

    def __str__(self):
        ret =  (f"win_fract:{self.win_fract}"
                f" x_min:{self.x_min}"
                f" y_min:{self.y_min}"
                f" x_max:{self.x_max}"
                f" y_max:{self.y_max}"
                f" ncols:{self.ncols}"
                f" nrows:{self.nrows}")
        return ret

 
class MagnifySelect:
    """ Selection info """
    def __init__(self, ix_min=None, iy_min=None,
                        ix_max=None, iy_max=None,):
        self.ix_min = ix_min
        self.iy_min = iy_min
        self.ix_max = ix_max
        self.iy_max = iy_max
        
    def __str__(self):
        ret =  (f"ix_min:{self.ix_min}"
                f" iy_min:{self.iy_min}"
                f" ix_max:{self.ix_max}"
                f" iy_max:{self.iy_max}")
        return ret
        
class MagnifyInfo:
    """ Magnification info
        For sending info from and to original canvas
    """
    info_number = 0         # Ascending unique number
                            # Bumped by __init__, make_child
    def __init__(self, select=None,
                 top_region=None,
                 display_region=None,
                 description="",
                 display_window=None,
                 parent_info=None,
                 child_infos=[],
                 mag_nrows=None,
                 mag_ncols=None,
                 base_canvas=None
                 ):
        """
        Setup magnification information, to be modified
        :select: (MagnifySelect) selection cell indexes
        :top_region: (MagnifyDisplayRegion) - top display region
        :display_region: (MagnifyDisplayRegion) display region on which selection is made
        :display_window: (AudioDrawWindow) - display window
        :parent_info: (MagnifyInfo) - display from which this info was selected
        :child_infos: (list of MagnifyInfo) - children from this info
        :mag_nrows: magnified number of rows default: display_region.nrows 
        :mag_ncols: magnified number of columns default: display_region.nrows 
        :base_canvas: canvas with base info/control
        """
        MagnifyInfo.info_number += 1
        self.info_number = MagnifyInfo.info_number
        if select is None:
            select = MagnifySelect()
        self.select = select
        if top_region is None:
            top_region = MagnifyDisplayRegion()
        self.top_region = top_region
        if display_region is None:
            display_region = MagnifyDisplayRegion(dup=self.top_region)
        self.display_region = display_region
        self.description = description
        if mag_nrows is None:
            mag_nrows = display_region.nrows
            if mag_nrows is None:
                mag_nrows = top_region.nrows
        self.mag_nrows = mag_nrows
        if mag_ncols is None:
            mag_ncols = display_region.ncols
            if mag_ncols is None:
                mag_ncols = top_region.ncols
        self.mag_ncols = mag_ncols
        self.parent_info = parent_info
        self.child_infos = child_infos
        self.base_canvas = base_canvas

    def __str__(self):
        ret = f"MagI[{self.info_number}]:"
        if self.description:
            ret += f" self.description"
        if self.select is not None:
            ret += f" select: {self.select}"
        if self.top_region is not None:
            ret += f"\n    top:{self.top_region}"
        if self.display_region is not None:
            ret += f"\n    display:{self.display_region}"
        if self.mag_ncols != self.display_region.ncols:
            ret += f" mag_ncols:{self.mag_ncols}"    
        if self.mag_nrows != self.display_region.nrows:
            ret += f" mag_ncols:{self.mag_nrows}"
        if self.parent_info is None:
            ret += f"\n    no parent"
        else:
            np = 0      # count levels
            parent = self.parent_info
            while parent is not None:
                np += 1
                parent = parent.parent_info
            ret += f"\n parent level: {np}"
        nc = len(self.child_infos)
        ret += f"\n    {nc}"
        ret += " child" if nc == 1 else " children"
        return ret

    def make_child(self):
        """ Create child info with us as parent and child added to children
        :returns: (MagnifyInfo) child
        """
        child = copy.copy(self)     # Don't expect changing except as below
        MagnifyInfo.info_number += 1
        child.info_number = MagnifyInfo.info_number
        child.description = ""
        child.select = MagnifySelect()  # Initialize as unselected
        child.display_window = None     # Set when used
        child.parent_info = self
        child.child_infos = []         
        self.child_infos.append(child)
        return child
        

if __name__ == '__main__':
    mag_info = MagnifyInfo()
    print(f"mag_info: {mag_info}")        


#grid_rot.py    23Nov2022  crs,
"""
Grid Rotation - handling discrete 3by3 square rotation
Used iby:
 BrailleWindow, AudioWindow TurtleBraille,BralleDisplay
"""

class GridRot:
    # incremental directions
    RIGHT = 2
    LEFT = 4
    UP = 2
    DOWN = 6

    """
    Integral direction 0-7
         (xchg,ychg), by irot 
        | 3(-1,1)           2(0,1)           1(1,1)     |
        |                                               |
        | 4(-1,0)                            0(1,0)     |
        |                                               |
        | -3/5(-1,-1)      -2/6(0,-1)       -1/7(1,-1)  |
           
    e.g. 0(right) with irot=1 gives 1(1,1)(diag up right)
    
    """
    #        0        1       2      3       4
    idirs = [(1,0),
                     (1,1),   (0,1), (-1,1),
                                             (-1,0),
                     (-1,-1), (0,-1), (1,-1)]
    #                 -3       -2      -1      

    
    @classmethod
    def irot_to_ixy(cls, irot):
        """ Convert rotation int to {ixchg,iyychg)
        :irot: integer rotation pos counter clockwise
        :returns: (ixchg, iychg)
        """
        return cls.idirs[len(cls.idirs)%irot]

    @classmethod
    def ixy_to_irot(cls, ixy):
        """ Convert {ixchg,iyychg) to rotation int
        :ixy: integer rotation pos counter clockwise
        :returns: irot integer rotation
        """
        irot = cls.idirs.find(ixy)
        if irot > len(cls.idirs)//2:
            irot = irot - len(cls.idirs)
        return irot

    @classmethod
    def next_cell(cls, ixy, idir=None, dist=1):
        """ Get the next cell
        :ixy: cell tuple on which begin
        :idir: direction index in GR.idirs(ix_change, iy_change)
            default: GR.Up 
        :dist: change in direction i.e., dist 1 for adjacent
               cell default: 1
        :returns: (ix,iy) new cell tuple
        """
        if idir is None:
            idir = GridRot.UP
        dir = cls.idirs[idir%len(cls.idirs)]
        ixy_new = (ixy[0]+dir[0]*dist,
                   ixy[1]+dir[1]*dist)
        return ixy_new
    
    @classmethod
    def dir_irot(cls, idir, irot):
        """ Integral rotation of dir (xchg,ychg), by irot 
            | 3(-1,1)        2(0,1)         1(1,1)   |
            |                                        |
            | 4(-1,0)                       0(1,0)   |
            |                                        |
            | -3(-1,-1)     -2(0,-1)       -1(1,-1)  |
               
        e.g. 0(right) with irot=1 gives 1(1,1)(diag up right)
        
        """
        idir_rot = len(cls.idirs)%(idir+irot)
        return idir_rot



#sinewave_numpy.py    29Mar2023  crs
""" Lower level substitute for pysinewave
    Not a direct substitute but for lower level operations
"""
import time
import copy
import numpy as np
import sounddevice as sd
from pysinewave import utilities

from select_trace import SlTrace

class SineWaveNumPy:
    """ Gathers and plays stereo sine wave
    """
    @classmethod
    def concatinate(cls, sinewave_nps):
        """ Concatinate a list of SineWaveNumPy wave forms
        :sinewave_nps: List of SineWaveNumPy waves
        :returns: SineWaveNumPy waveform
        """
        sample_rate = sinewave_nps[0].sample_rate       # Use first entry
        duration = sinewave_nps[0].duration 
        wfs_ndarr = [swnp.wf_ndarr for swnp in sinewave_nps]  # Get the waveforms
        wfc_ndarr = np.concatenate(wfs_ndarr)
        swnp = SineWaveNumPy(wf_ndarr=wfc_ndarr, sample_rate=sample_rate,
                           duration=duration)
        return swnp
        
    def __init__(self, pitch=0, decibels_left=0, decibels_right=0,
                sample_rate=44100, duration=None, delay=None,
                wf_ndarr=None):
        """ Setup waves
        :pitch: user tone level default: 0
        :decibels_left: left volume in decibels default: 0
        :decibels_right: right volume in decibels default: 0
        :samplerate: samples per second default: 44100
        :duration: play duration(seconds) default:calculated from len, sample_rate
        :wf_ndarr: if present, BYPASS calculation and use as waveform (ndarray)
        """
        if wf_ndarr is not None:
            self.sample_rate = sample_rate
            self.duration = duration
            self.wf_ndarr = wf_ndarr
            self.delay = delay
            
        else:
            freq_hz = utilities.pitch_to_frequency(pitch)
            atten_left = utilities.decibels_to_amplitude_ratio(decibels_left) 
            atten_right = utilities.decibels_to_amplitude_ratio(decibels_right) 
            self.sample_rate = sample_rate
            if duration is None:
                SlTrace.lg(f"SineWaveNumPy duration is {duration} treat as .1")
                duration = .1
            self.duration = duration
            self.delay = delay
            # NumpPy magic to calculate the waveform
            each_sample_number = np.arange(duration * sample_rate)
            base_waveform = np.sin(2 * np.pi * each_sample_number * freq_hz / sample_rate)
            left_waveform = base_waveform.reshape(-1,1)*atten_left
            right_waveform = base_waveform.reshape(-1,1)*atten_right
            wf_ndarr = np.hstack((left_waveform, right_waveform))
            self.wf_ndarr = copy.deepcopy(wf_ndarr)
        
    def play(self):
        """ Start playing tone
        """
        sd.play(self.wf_ndarr, self.sample_rate)
        
        
    def stop(self):
        """ Stop playing tone
        """
        sd.stop()

if __name__ == "__main__":
    pitch = 2
    pitch_sep = 2
    decibels_sep = 50
    decibels_left = -50
    decibels_right = decibels_left + decibels_sep
    duration_s = 2
    sw1 = SineWaveNumPy(pitch=pitch, decibels_left=decibels_left,
                       decibels_right=decibels_right,
                       duration_s=duration_s)
    sw2 = SineWaveNumPy(pitch=pitch+pitch_sep, decibels_left=decibels_right,
                       decibels_right=decibels_left,
                       duration_s=duration_s)
    print(f"Starting with pitch:{pitch} duration:{duration_s} sec")
    sw1.play()
    time.sleep(duration_s)
    sw1.stop()
    print("End of sw1")
    time.sleep(1)
    sw2.play()
    time.sleep(duration_s)
    sw2.stop()
    print("End of sw2")


# wx_play_sound_control    31Mar2023  crs
"""
Control sound playing
Soon to be directed to SpeakerControler
"""
from select_trace import SlTrace

from play_sound import PlaySound

class PlaySoundControl:
    
    def __init__(self, dur=.1, volume=None,):
        """ Setup from sinewave beep
        :swb: SineWaveBeep instance
        """
        self.dur = dur
        self.volume = volume
        self.g_psid = 0          # Unique sound id generator
        self.play_sound_running = {}     # currently playing by psid
        self.play_sound_available = {}   # free/stopped by pskey - avail lists

    def new_id(self):
        """ Generate unique sound id
        :returns: new id
        """
        self.g_psid += 1
        return self.g_psid

    def play_sound(self,pitch, dur=None, volume=None, dly=None):
        """ Play sound - non-blocking
        :pitch: sound pitch 
        :dur: sound duration default:dur_base (msec)
        :volume: sound volume in decibels default: volume_base
        :dly: delay(silence) before sound
        """
        self.update()
        if volume is None:
            volume = self.volume
        SlTrace.lg(f"play_sound:pitch:{pitch} dur:{dur} volume:{volume} ", "pos_tracking")
        ps = self.get_sound(pitch, volume=volume)
        SlTrace.lg(f"    ps:{ps}", "sound")
        self.add_ps_to_run(ps)        # Show who's running
        ps.play(dly=dly, dur=dur)

    def play_sounds(self,pitches, dur=None, volume=None, dly=None):
        """ Play sound - non-blocking
        :pitches: list of sound pitch 
        :dur: sound duration default:dur_base (msec)
        :volume: sound volume in decibels default: volume_base
        :dly: delay(silence) before sound
        """
        self.update()
        SlTrace.lg(f"play_sounds:pitches:{pitches} dur:{dur} volume:{volume} ", "sound")
        ps_list = []
        for pitch in pitches:
            ps = self.get_sound(pitch, volume=volume)
            SlTrace.lg(f"    ps:{ps}", "sound")
            ps_list.append(ps)
        for ps in ps_list:
            self.add_ps_to_run(ps)        # Show who's running
            ps.play(dly=dly, dur=dur)

    def add_ps_to_run(self, ps):
        """ Add PlaySound instance to runing list
        :ps: PlaySound instance
        """
        if ps.psid in self.play_sound_running:
            SlTrace.lg(f"Attempted adding of running PlaySound[{ps.psid}] - {ps}")
            return
        
        self.play_sound_running[ps.psid] = ps
        

    def free_ps(self, ps):
        """ Free PlaySound resource for possible reuse
        :ps: PlaySound instance
        """
        self.update_pending()
        if ps.psid not in self.play_sound_running:
            SlTrace.lg(f"Attempted freeing of non-running PlaySound[{ps.psid}] - {ps}")
            return
        
        if ps.pskey in self.play_sound_available:
            psavails = self.play_sound_available[ps.pskey]
        else:
            psavails = self.play_sound_available[ps.pskey] = []
        psavails.append(ps)
        del(self.play_sound_running[ps.psid])
         
        
    def get_sound(self, pitch, volume=None):
        """ get or create PlaySound instance
        :pitch, volume: see PlaySound
        :returns: PlaySound 
        """
        self.update_pending()
        if volume is None:
            volume = self.volume_base

        pskey = PlaySound.pskey(pitch=pitch,volume=volume)
        ps = PlaySound(self, pitch=pitch,
                          volume=volume, pskey=pskey)
        return ps

        ####### Don't reuse #######
         
        if pskey in self.play_sound_available:
            psavails = self.play_sound_available[pskey]
            psavail = psavails.pop()
            SlTrace.lg(f"get_sound reusing: {psavail}", "sound")
            if len(psavails) == 0:
                del(self.play_sound_available[pskey]) 
            return psavail
        
        ps = PlaySound(self, pitch=pitch,
                          volume=volume, pskey=pskey)
        return ps

    def update(self):
        """ tkinter update
        """
        ###wxport###
        
    def update_pending(self):
        """ Update pending events
        """
        ###wxport###



from __future__ import annotations

import logging
import os
import re
import site
import sys

logger = logging.getLogger(__name__)
_INCLUDE_SYSTEM_SITE_PACKAGES_REGEX = re.compile(
    r"include-system-site-packages\s*=\s*(?P<value>true|false)"
)


def _running_under_venv() -> bool:
    """Checks if sys.base_prefix and sys.prefix match.

    This handles PEP 405 compliant virtual environments.
    """
    return sys.prefix != getattr(sys, "base_prefix", sys.prefix)


def _running_under_legacy_virtualenv() -> bool:
    """Checks if sys.real_prefix is set.

    This handles virtual environments created with pypa's virtualenv.
    """
    # pypa/virtualenv case
    return hasattr(sys, "real_prefix")


def running_under_virtualenv() -> bool:
    """True if we're running inside a virtual environment, False otherwise."""
    return _running_under_venv() or _running_under_legacy_virtualenv()


def _get_pyvenv_cfg_lines() -> list[str] | None:
    """Reads {sys.prefix}/pyvenv.cfg and returns its contents as list of lines

    Returns None, if it could not read/access the file.
    """
    pyvenv_cfg_file = os.path.join(sys.prefix, "pyvenv.cfg")
    try:
        # Although PEP 405 does not specify, the built-in venv module always
        # writes with UTF-8. (pypa/pip#8717)
        with open(pyvenv_cfg_file, encoding="utf-8") as f:
            return f.read().splitlines()  # avoids trailing newlines
    except OSError:
        return None


def _no_global_under_venv() -> bool:
    """Check `{sys.prefix}/pyvenv.cfg` for system site-packages inclusion

    PEP 405 specifies that when system site-packages are not supposed to be
    visible from a virtual environment, `pyvenv.cfg` must contain the following
    line:

        include-system-site-packages = false

    Additionally, log a warning if accessing the file fails.
    """
    cfg_lines = _get_pyvenv_cfg_lines()
    if cfg_lines is None:
        # We're not in a "sane" venv, so assume there is no system
        # site-packages access (since that's PEP 405's default state).
        logger.warning(
            "Could not access 'pyvenv.cfg' despite a virtual environment "
            "being active. Assuming global site-packages is not accessible "
            "in this environment."
        )
        return True

    for line in cfg_lines:
        match = _INCLUDE_SYSTEM_SITE_PACKAGES_REGEX.match(line)
        if match is not None and match.group("value") == "false":
            return True
    return False


def _no_global_under_legacy_virtualenv() -> bool:
    """Check if "no-global-site-packages.txt" exists beside site.py

    This mirrors logic in pypa/virtualenv for determining whether system
    site-packages are visible in the virtual environment.
    """
    site_mod_dir = os.path.dirname(os.path.abspath(site.__file__))
    no_global_site_packages_file = os.path.join(
        site_mod_dir,
        "no-global-site-packages.txt",
    )
    return os.path.exists(no_global_site_packages_file)


def virtualenv_no_global() -> bool:
    """Returns a boolean, whether running in venv with no system site-packages."""
    # PEP 405 compliance needs to be checked first since virtualenv >=20 would
    # return True for both checks, but is only able to use the PEP 405 config.
    if _running_under_venv():
        return _no_global_under_venv()

    if _running_under_legacy_virtualenv():
        return _no_global_under_legacy_virtualenv()

    return False



# adw_menus    09Mar2023  crs, Split off from adw_front_end

import tkinter as tk

from select_trace import SlTrace
from trace_control_window import TraceControlWindow


class MenuDisp:
    """ Menu dispatch table entry
    Supporting multiple mode dispatch (e.g, Dropdown item plus command mode)
    """

    def __init__(self, label, command, underline):
        self.label = label
        self.command = command
        self.underline = underline
        self.shortcut = label[underline].lower()


class AdwMenus:
    def __init__(self, adw_front_end):
        """ Setup menus for AudioDrawWindow
        :adw_front_end: (AdwFrontEnd)
        """
        self.fte = adw_front_end
        self.mw = adw_front_end.mw

        self.menu_setup()

    def on_alt_a(self, event):
        """ keep from key-press
        """
        SlTrace.lg("on_alt_a")

    def on_alt_m(self, event):
        """ keep from key-press
        """
        SlTrace.lg("on_alt_m")

    def on_alt_n(self, event):
        """ keep from key-press
        """
        SlTrace.lg("on_alt_n")

    def on_alt_f(self, event):
        """ keep from key-press
        """
        SlTrace.lg("on_alt_f")

    def on_alt_s(self, _):
        """ keep from key-press
        """
        SlTrace.lg("on_alt_s")

    def menu_setup(self):
        """ Setup menu system
        """
        # creating a menu instance
        self.mw.bind('<Alt-a>', self.on_alt_f)  # Keep this from key cmds
        self.mw.bind('<Alt-d>', self.on_alt_f)  # Keep this from key cmds
        self.mw.bind('<Alt-f>', self.on_alt_f)  # Keep this from key cmds
        self.mw.bind('<Alt-n>', self.on_alt_n)  # Keep this from key cmds
        self.mw.bind('<Alt-M>', self.on_alt_m)  # Keep this from key cmds
        self.mw.bind('<Alt-m>', self.on_alt_m)  # Keep this from key cmds
        self.mw.bind('<Alt-N>', self.on_alt_n)  # Keep this from key cmds
        self.mw.bind('<Alt-n>', self.on_alt_n)  # Keep this from key cmds
        self.mw.bind('<Alt-S>', self.on_alt_s)  # Keep this from key cmds
        self.mw.bind('<Alt-s>', self.on_alt_s)  # Keep this from key cmds

        menubar = tk.Menu(self.mw)
        self.menubar = menubar  # Save for future reference
        self.mw.config(menu=menubar)

        self.Properties = None
        self.LogFile = None

        file_menu = tk.Menu(menubar, tearoff=0)
        self.file_menu_setup(file_menu)
        menubar.add_cascade(label="File", menu=file_menu)

        mag_menu = tk.Menu(menubar, tearoff=0)
        self.mag_menu_setup(mag_menu)
        menubar.add_cascade(label="Magnify", menu=mag_menu)

        nav_menu = tk.Menu(menubar, tearoff=0)
        self.nav_menu_setup(nav_menu)
        menubar.add_cascade(label="Navigate", menu=nav_menu)

        draw_menu = tk.Menu(menubar, tearoff=0)
        self.draw_menu_setup(draw_menu)
        menubar.add_cascade(label="Draw", menu=draw_menu)

        scan_menu = tk.Menu(menubar, tearoff=0)
        self.scan_menu_setup(scan_menu)
        menubar.add_cascade(label="Scanning", menu=scan_menu)

        aux_menu = tk.Menu(menubar, tearoff=0)
        aux_menu.add_command(label="Trace", command=self.trace_menu,
                             underline=0)
        menubar.add_cascade(label="Auxiliary", menu=aux_menu)

    """ File Menu setup package
    """

    def file_menu_setup(self, file_menu):
        self.file_menu = file_menu
        self.file_dispatch = {}
        self.file_menu_add_command(label="Open", command=self.File_Open_tbd,
                                   underline=0)
        self.file_menu_add_command(label="Save", command=self.File_Save_tbd,
                                   underline=0)
        file_menu.add_separator()
        self.file_menu_add_command(label="Log", command=self.LogFile,
                                   underline=0)
        self.file_menu_add_command(label="Properties", command=self.Properties,
                                   underline=0)
        file_menu.add_separator()
        self.file_menu_add_command(label="Exit", command=self.pgm_exit,
                                   underline=1)

    def file_menu_add_command(self, label, command, underline):
        """ Setup menu commands, setup dispatch for direct call
        :label: add_command label
        :command: command to call
        :underline: short-cut index in label
        """
        self.file_menu.add_command(label=label, command=command,
                                   underline=underline)
        menu_de = MenuDisp(label=label, command=command,
                           underline=underline)

        self.file_dispatch[menu_de.shortcut] = menu_de

    def file_direct_call(self, short_cut):
        """ Short-cut call direct to option
        :short_cut: one letter option for call 
        """
        if short_cut not in self.file_dispatch:
            raise Exception(f"file option:{short_cut} not recognized")
        menu_de = self.file_dispatch[short_cut]
        menu_de.command()

    def File_Open_tbd(self):
        print("File_Open_menu to be determined")

    def File_Save_tbd(self):
        print("File_Save_menu to be determined")

    """ draw_menu setup
    """

    def draw_menu_setup(self, draw_menu):
        self.draw_menu = draw_menu
        self.draw_dispatch = {}
        self.draw_menu_add_command(label="Help", command=self.draw_help,
                                   underline=0)
        self.draw_menu_add_command(label="drawing", command=self.start_drawing,
                                   underline=0)
        self.draw_menu_add_command(label="stop_drawing", command=self.stop_drawing,
                                   underline=0)

    def draw_menu_add_command(self, label, command, underline):
        """ Setup menu commands, setup dispatch for direct call
        :label: add_command label
        :command: command to call
        :underline: short-cut index in label
        """
        self.draw_menu.add_command(label=label, command=command,
                                   underline=underline)
        menu_de = MenuDisp(label=label, command=command,
                           underline=underline)

        self.draw_dispatch[menu_de.shortcut] = menu_de

    def draw_direct_call(self, short_cut):
        """ Short-cut call direct to option
        :short_cut: one letter option for call 
        """
        if short_cut not in self.draw_dispatch:
            raise Exception(f"draw option:{short_cut} not recognized")
        menu_de = self.draw_dispatch[short_cut]
        menu_de.command()

    def draw_help(self):
        """ Help for drawing
        """
        """ Help - list command (Alt-d) commands
        """
        help_str = """
        Help - list drawing setup commands (Alt-d) commands
        h - say this help message
        d - Start/enable drawing
        s - stop/disable drawing
        Escape - flush pending report output
        """
        self.speak_text(help_str)

    """ Magnify support package
    """

    """ Magnify menu commands  """

    def mag_help(self):
        """ Help for Alt-m commands
        """
        """ Help - list command (Alt-m) commands
        """
        help_str = """
        Help - list magnify commands (Alt-m) commands
        h - say this help message
        t - expand magnify selected region left/right
        s - select/mark magnify region
        t - expand magnify selected region up/down top/bottom
        v - view region (make new window)
        """
        self.speak_text(help_str)

    def mag_menu_setup(self, mag_menu):
        self.mag_menu = mag_menu
        self.mag_dispatch = {}
        self.mag_menu_add_command(label="Help", command=self.mag_help,
                                  underline=0)
        self.mag_menu_add_command(label="Remove Pos History", command=self.erase_pos_history,
                                  underline=0)
        self.mag_menu_add_command(label="Select", command=self.mag_select,
                                  underline=0)
        self.mag_menu_add_command(label="Expand Right", command=self.mag_expand_right,
                                  underline=7)
        self.mag_menu_add_command(label="Expand Top", command=self.mag_expand_top,
                                  underline=7)
        self.mag_menu_add_command(label="View", command=self.mag_view,
                                  underline=0)

    def mag_menu_add_command(self, label, command, underline):
        """ Setup menu commands, setup dispatch for direct call
        :label: add_command label
        :command: command to call
        :underline: short-cut index in label
        """
        self.mag_menu.add_command(label=label, command=command,
                                  underline=underline)
        menu_de = MenuDisp(label=label, command=command,
                           underline=underline)

        self.mag_dispatch[menu_de.shortcut] = menu_de

    def mag_direct_call(self, short_cut):
        """ Short-cut call direct to option
        :short_cut: one letter option for call 
        """
        if short_cut not in self.mag_dispatch:
            raise Exception(f"mag option:{short_cut} not recognized")
        magde = self.mag_dispatch[short_cut]
        magde.command()

    """ End of Magnify support
    """

    """ Scanning support package
    """

    """ Scanning menu commands  """

    def scan_help(self):
        """ Help for Alt-s commands
        """
        """ Help - list command (Alt-s) commands
        """
        help_str = """
        Help - list scanning commands (Alt-s) commands
        h - say this help message
        c - combine wave - faster scan
        d - disable combine wave - viewing in window
        k - toggle skip space
        s - Start scanning mode
        t - Stop scanning mode
        n - no_item_wait
        w - wait for items
        """
        self.speak_text(help_str)

    def scan_menu_setup(self, scan_menu):
        self.scan_menu = scan_menu
        self.scan_dispatch = {}
        self.scan_menu_add_command(label="Help", command=self.scan_help,
                                   underline=0)
        self.scan_menu_add_command(label="c -  combine wave - faster scan",
                                   command=self.scan_combine_wave,
                                   underline=0)
        self.scan_menu_add_command(label="d - disable combine wave",
                                   command=self.scan_disable_combine_wave,
                                   underline=0)
        self.scan_menu_add_command(label="f - flip skip space", command=self.flip_skip_space,
                                   underline=0)
        self.scan_menu_add_command(label="r - flip skip run", command=self.flip_skip_run,
                                   underline=0)
        self.scan_menu_add_command(label="Start scanning", command=self.start_scanning,
                                   underline=0)
        self.scan_menu_add_command(label="Stop scanning", command=self.stop_scanning,
                                   underline=1)
        self.scan_menu_add_command(label="No item wait", command=self.scan_no_item_wait,
                                   underline=0)
        self.scan_menu_add_command(label="Wait for item", command=self.scan_item_wait,
                                   underline=0)

    def scan_menu_add_command(self, label, command, underline):
        """ Setup menu commands, setup dispatch for direct call
        :label: add_command label
        :command: command to call
        :underline: short-cut index in label
        """
        self.scan_menu.add_command(label=label, command=command,
                                   underline=underline)
        menu_de = MenuDisp(label=label, command=command,
                           underline=underline)

        self.scan_dispatch[menu_de.shortcut] = menu_de

    def scan_direct_call(self, short_cut):
        """ Short-cut call direct to option
        :short_cut: one letter option for call 
        """
        if short_cut not in self.scan_dispatch:
            raise Exception(f"scan option:{short_cut} not recognized")
        scande = self.scan_dispatch[short_cut]
        scande.command()

    def scan_combine_wave(self):
        self.set_combine_wave()

    def scan_disable_combine_wave(self):
        self.set_combine_wave(val=False)

    def scan_no_item_wait(self):
        self.set_no_item_wait()

    def scan_item_wait(self):
        """clear no item wait
        """
        self.set_no_item_wait(val=False)

    """ End of Scanning support
    """

    """ Navigate support package
    """

    def nav_help(self):
        """ Help for Alt-n commands
        """
        """ Help - list command (Alt-n) commands
        """
        help_str = """
        Help - list navigate commands (Alt-n) commands
        h - say this help message
        a - Start reporting position
        b - remove 
        z - stop reporting position
        e - echo input on
        o - echo off
        v - visible cells
        i - invisible cells
        r - redraw figure
        s - silent speech
        t - talking speech
        l - log speech
        m - show marked(even if invisible)
        n - no log speech
        p - report position
        u - audio beep
        d - no audio beep
         Escape - flush pending report output
        """
        self.speak_text(help_str)

    def nav_menu_setup(self, nav_menu):
        self.nav_menu = nav_menu
        self.nav_dispatch = {}
        self.nav_menu_add_command(label="Help", command=self.nav_help,
                                  underline=0)
        self.nav_menu.add_command(label="add At loc", command=self.nav_add_loc,
                                  underline=0)
        self.nav_menu_add_command(label="b-remove At loc", command=self.nav_no_add_loc,
                                  underline=0)
        self.nav_menu_add_command(label="echo input on", command=self.nav_echo_on,
                                  underline=0)
        self.nav_menu_add_command(label="echo off", command=self.nav_echo_off,
                                  underline=5)

        self.nav_menu_add_command(label="visible cells", command=self.nav_make_visible,
                                  underline=0)
        self.nav_menu_add_command(label="invisible cells", command=self.nav_make_invisible,
                                  underline=0)
        self.nav_menu_add_command(label="marked", command=self.nav_show_marked,
                                  underline=0)
        self.nav_menu_add_command(label="noisy", command=self.make_noisy,
                                  underline=0)
        self.nav_menu_add_command(label="silent", command=self.make_silent,
                                  underline=0)
        self.nav_menu_add_command(label="talking", command=self.nav_make_talk,
                                  underline=0)
        self.nav_menu_add_command(label="log talk", command=self.nav_logt,
                                  underline=0)
        self.nav_menu_add_command(label="no log talk", command=self.nav_no_logt,
                                  underline=10)
        self.nav_menu_add_command(label="position", command=self.nav_say_position,
                                  underline=0)
        self.nav_menu_add_command(label="redraw figure", command=self.nav_redraw,
                                  underline=2)
        self.nav_menu_add_command(label="audio beep", command=self.nav_audio_beep,
                                  underline=1)
        self.nav_menu_add_command(label="q no audio beep",
                                  command=self.nav_no_audio_beep,
                                  underline=0)
        self.nav_menu_add_command(label="x enable mouse navigation",
                                  command=self.nav_enable_mouse,
                                  underline=0)
        self.nav_menu_add_command(label="y disable mouse navigation",
                                  command=self.nav_disable_mouse,
                                  underline=0)

    def nav_menu_add_command(self, label, command, underline):
        """ Setup menu commands, setup dispatch for direct call
        :label: add_command label
        :command: command to call
        :underline: short-cut index in label
        """
        self.nav_menu.add_command(label=label, command=command,
                                  underline=underline)
        menu_de = MenuDisp(label=label, command=command,
                           underline=underline)

        self.nav_dispatch[menu_de.shortcut] = menu_de

    def nav_direct_call(self, short_cut):
        """ Short-cut call direct to option
        :short_cut: one letter option for call 
        """
        if short_cut not in self.nav_dispatch:
            raise Exception(f"nav option:{short_cut} not recognized")
        navde = self.nav_dispatch[short_cut]
        navde.command()

    """
    Trace support
    """

    def trace_menu(self):
        TraceControlWindow(tcbase=self.mw)

    """
    Links to front end 
    """
    """
    General
    """

    def erase_pos_history(self):
        self.fte.erase_pos_history()

    def speak_text(self, msg, dup_stdout=True,
                   msg_type=None,
                   rate=None, volume=None):
        """ Speak text, if possible else write to stdout
        :msg: text message, iff speech
        :dup_stdout: duplicate to stdout default: True
        :msg_type: type of speech default: 'REPORT'
            REPORT - standard reporting
            CMD    - command
            ECHO - echo user input
        :rate: speech rate words per minute
                default: 240
        :volume: volume default: .9
            
        """
        self.fte.speak_text(msg=msg, msg_type=msg_type,
                            dup_stdout=dup_stdout,
                            rate=rate, volume=volume)

    """
     File links
      most local
    """

    def pgm_exit(self):
        self.fte.pgm_exit()

    """
    Magnification links
    """

    def mag_select(self):
        """ Select magnification region
            -rectangle including all figure cells traveled so far
            :returns: True if some selected else False
        """
        return self.fte.mag_select()

    def mag_expand_right(self):
        """ Expand selection region right and left by 20%
        """
        self.fte.mag_exapnd_right()

    def mag_expand_top(self):
        """ Expand selection region top/bottom by 20%
        """
        self.fte.mag_exapnd_top()

    def mag_view(self):
        """ View selected region, creating a new AudioDrawWindow
        """
        self.fte.mag_view()

    """ drawing
    """

    def start_drawing(self):
        """ Start/enable drawing
        """
        self.fte.start_drawing()

    def stop_drawing(self):
        """ Stop/disable drawing
        """

    """ End of Magnification links """

    """
     Navigation / Front End links
     """

    def set_enable_mouse(self, val=True):
        """ Enable/Disable mouse dragging operation
        :val: value to set default: True - enable
        """
        self.fte.set_enable_mouse(val=val)

    def is_enable_mouse(self):
        """ Check if mouse dragging enabled
        """
        return self.fte.is_enable_mouse()

    def nav_add_loc(self):
        self.fte.nav_add_loc()

    def nav_no_add_loc(self):
        self.fte.nav_no_add_loc()

    def nav_echo_on(self):
        self.fte.nav_echo_on()

    def nav_echo_off(self):
        self.fte.nav_echo_off()

    def nav_make_visible(self):
        self.fte.nav_make_visible()

    def nav_make_invisible(self):
        self.fte.nav_make_invisible()

    def nav_show_marked(self):
        self.fte.nav_show_marked()

    def make_noisy(self):
        self.fte.make_noisy()

    def make_silent(self):
        self.fte.make_silent()

    def nav_make_talk(self):
        self.fte.nav_make_talk()

    def nav_logt(self):
        self.fte.nav_logt()

    def nav_no_logt(self):
        self.fte.nav_no_logt()

    def nav_say_position(self):
        self.fte.nav_say_position()

    def nav_redraw(self):
        self.fte.nav_redraw()

    def nav_audio_beep(self):
        self.fte.nav_audio_beep()

    def nav_no_audio_beep(self):
        self.fte.nav_no_audio_beep()

    def nav_enable_mouse(self):
        self.set_enable_mouse()

    def nav_disable_mouse(self):
        self.set_enable_mouse(False)

    """ End of Navigation links """

    """
    Scanning support
        through fte
    """

    def set_combine_wave(self, val=True):
        """ Enable/disable combine wave scanning mode
        :val: value for mode
        """
        self.fte.set_combine_wave(val=val)

    def set_no_item_wait(self, val=True):
        """ Set/clear scanning no_wait option
        :val: True - no waiting
        """
        self.fte.set_no_item_wait(val=val)

    def flip_skip_space(self):
        """ Flip skipping spaces
        """
        self.fte.flip_skip_space()

    def flip_skip_run(self):
        """ Flip skipping run of equals
        """
        self.fte.flip_skip_run()

    def start_scanning(self):
        self.fte.start_scanning()

    def stop_scanning(self):
        self.fte.stop_scanning()



#wx_tk_rpc_host.py   24Jan2024  crs, renamed from wx_tk_rem_access.py
""" Remote access to tk canvas information
tkinter canvas resides in host process
wxPython AudioDrawWindow instance(s) reside in client
processing is done in threads to avoid locking
"""
import tkinter as tk
import socket as sk
#from wx_tk_rpc_server import TkRPCServer
import queue
import pickle
import time
import threading as th
import copy

#from wx_rpc import RPCServer
from wx_bg_rpc import RPCServer
from wx_rpc import RPCClient

from wx_canvas_grid import CanvasGrid

from select_trace import SlTrace
from wx_tk_bg_call import TkBgCall

class TkRPCHost:
    HOST_PORT = 50010
    """Host control containing tkinter canvas
    using socket TCP/IP
    """
    
    def __init__(self, canvas_grid, root=None, host_name='localhost',
                 host_port=None,
                 cmd_time_ms= 500,
                 max_recv=2**16):
        """ Setup cmd interface
        :canvas_grid: tkinter CanvasGrid
        :root: tkinter main window, used for update() calls
        :host_name: name/id default: '' this machine
        :host_port: port number where we accept unsolicited input
                        default: 50020
        :max_recv: maximum data received length in bytes
                    default: 2**16
        :cmd_time_ms: maximum between cmd check in seconds
                    default: .5 sec
        """
        self.sock_out = None    #  setup once connection is establised 
        self.canvas_grid = canvas_grid
        self.canvas_grid_snapshots = []     # Snapshots, if any
        self.root = root
        self.host_name = host_name
        if host_port is None:
            host_port = TkRPCHost.HOST_PORT
        self.host_port = host_port
        self.max_recv = max_recv
        self.cmd_time_ms = cmd_time_ms
        self.user_is_ready = False      # Set True when user is ready to accept calls
        self.snapshot_is_complete = True    # Set if no waiting required

        self.bg = TkBgCall(root, canvas_grid=self.canvas_grid,
                canvas_grid_snapshots=self.canvas_grid_snapshots)
        SlTrace.lg(f"self.to_host_server ="
                   f" TkRPCServer({self.host_name}, {self.host_port})", "HOST")
        self.to_host_server = RPCServer(self.bg, self.host_name, self.host_port)

        self.to_host_server.registerMethod(self.setup_calling_user)
        self.to_host_server.registerMethod(self.get_canvas_lims)
        self.to_host_server.registerMethod(self.get_cell_rect_tur)
        self.to_host_server.registerMethod(self.get_cell_specs)
        self.to_host_server.registerMethod(self.snapshot_complete)
        self.to_host_server.registerMethod(self.test_dummy)
        self.to_host_server.registerMethod(self.get_ret)
        th.Thread(target=self.cmd_in_th_proc).start()

        
        
        SlTrace.lg("TkRPCHost __init__()", "HOST")            

    def test_dummy(self, *args, **kwargs):
        """ dummy test function
        """
        SlTrace.lg(f"test_dummy({args}, {kwargs})", "HOST")
        return ""
                                  
    def cmd_in_th_proc(self):
        SlTrace.lg(f"HOST: server_host_th_proc", "HOST")
        self.to_host_server.run()

    def setup_calling_user(self, user_name=None, user_port=None):
        """ Setup calling user - indicate user is ready to be called
        :user_name: user name default: our host name
        :user_port: user port default: generated from input port
        """
        if user_name is None:
            user_name = self.host_name
        if user_port is None:
            user_port = self.host_port+1
        SlTrace.lg(f"self.from_host_client ="
                   f" RPCClient({user_name}, {user_port})", "HOST")
        self.from_host_client = RPCClient(user_name, user_port)
        self.from_host_client.connect()
        self.user_is_ready = True
        SlTrace.lg("User is ready to accept calls", "HOST")
    
    def wait_for_user(self):
        """ Wait till user is ready
        """
        while not self.user_is_ready:
            self.tk_event_check()
    
    def wait_for_snapshot(self):
        """ Wait till snapshot is ready
        """
        while not self.snapshot_complete:
            self.tk_event_check()
            
    def tk_event_check(self):
        """ Allow tk event processing
        """                        
        self.root.update_idletasks()
        self.root.update()
        
    def snapshot(self, title=None):
        """ Create display snapshot
        :title: display title
        """
        self.wait_for_user()
        self.wait_for_snapshot()    # incase double snapshot
        # must call from main thread
        self.snapshot_is_complete = False
        SlTrace.lg(f"self.from_host_client.snapshot({title})", "snapshot")
        ###snapshot = copy.deepcopy(self.canvas_grid)   # fails
        snapshot = self.canvas_grid.copy()
        snapshot_str = snapshot.canvas_show_items()
        sno = len(self.canvas_grid_snapshots)+1
        SlTrace.lg(f"\nsnapshot[{sno}]: {snapshot_str}", "snapshot")
        self.canvas_grid_snapshots.append(snapshot)
        self.from_host_client.snapshot(f"{sno}: {title}",
                        snapshot_num=sno)
        self.wait_for_snapshot()    # Block till completed

    def snapshot_complete(self):
        """ Signal snapshot process has completed
        """        
        self.snapshot_is_complete = True
        
        
    def if_attr(self, cmd_dt, attr_name):
        """ Return attr value from dict if present else None
        :cmd_dt: command dictionary
        :attr_name: name of attribute
        :returns: attr value from dict iff attr_name present
                    else None
        """
        ret = None      # returned if attr_name not present
        if attr_name in cmd_dt:
            ret = cmd_dt[attr_name]
        return ret

    def make_bg_call(self, call_name):
        """ Invoke call to be waited on
        """
        bg = TkBgCall(call_name)
        
        self.root.after(0, bg.args)

    def get_ret(self, call_num):
        """ Get return value from call if availaible, None if not
        :call_num: unique call number
        """
        return self.bg.get_ret(call_num)
            
    def get_cell_specs(self,
                        snapshot_num=None,
                        win_fract=None, 
                        x_min=None, y_min=None,
                        x_max=None, y_max=None,
                        n_cols=None, n_rows=None):

        cell_specs = self.canvas_grid.get_cell_specs()
        SlTrace.lg(f"\nTkRPCHost:canvas_grid.get_cell_specs cell_specs(): {cell_specs}", "HOST")
        if snapshot_num is None or snapshot_num == 0:
            canvas_grid = self.canvas_grid
        else:
            canvas_grid = self.canvas_grid_snapshots
        ret = canvas_grid.get_cell_specs(
                        win_fract=win_fract, 
                        x_min=x_min, y_min=y_min,
                        x_max=x_max, y_max=y_max,
                        n_cols=n_cols, n_rows=n_rows)
        SlTrace.lg(f"""nTkRPCHost: canvas_grid.get_cell_specs(win_fract={win_fract}, 
                        x_min={x_min}, y_min={y_min},
                        x_max={x_max}, y_max={y_max},
                        n_cols={n_cols}, n_rows={n_rows})
                    ret : {ret}
                    """, "cell_specs")        
        return ret
        #return cell_specs   # TFD return strait from grid
    '''        
    def get_cell_specs_set(self,
                        win_fract=None, 
                        x_min=None, y_min=None,
                        x_max=None, y_max=None,
                        n_cols=None, n_rows=None):
        call_num = self.bg.set_call(self.get_cell_specs,
                        win_fract=win_fract, 
                        x_min=x_min, y_min=y_min,
                        x_max=x_max, y_max=y_max,
                        n_cols=n_cols, n_rows=n_rows)
        ret = self.get_ret(call_num)
        return ret
    '''    

    def get_cell_rect_tur(self,
                        ix=None, 
                        iy=None):

        ret = self.canvas_grid.get_cell_rect_tur(
                        ix,iy)
        SlTrace.lg(f"\nTkRPCHost:get_cell_rect_tur({ix},{iy}) ret: {ret}", "cell_specs")
        return ret     # TFD
            

    def get_canvas_lims(self):
        """ Get canvas limits - internal values, to which
        self.base.find_overlapping(cx1,cy1,cx2,cy2) obeys.
        NOTE: These values, despite some vague documentation, may be negative
              to adhere to turtle coordinate settings.
        
        :returns: internal (xmin, xmax, ymin, ymax)
        """
        ret = self.canvas_grid.get_canvas_lims()
        return ret
        
    def test_command(self, message="No Message Sent"):
        """ Do server test command
        :message: test message default: "No Message Sent"
        :returns: Answer: <message>
        """
        return "Answer: " + message       

if __name__ == '__main__':

    import tkinter as tk
    root = tk.Tk()

    cvg = CanvasGrid(root, height=450, width=450)
    cvg.create_line(0,0,200,300, width=10, fill="blue", tags="blue_tag")
    cvg.show_canvas()
    cvg.create_line(200,300,400,400, width=10, fill="green", tags=["green1","green2"])
    cvg.create_rectangle(200,200,300,300, fill="red")
    cvg.create_oval(100,200,250,300, fill="orange", tags="orange_tag")
    port = None
    tkh = TkRPCHost(cvg)
    cell_specs = tkh.get_cell_specs()
    SlTrace.lg(f"cell_specs: {cell_specs}", "cell_specs")
    '''
    '''
    root.mainloop()
    


# wx_canvas_panel_item.py   29Nov2023  crs, split off
"""
CanvasPanel item which is used to create/recreate display
"""
import wx
from wx_stuff import *

from select_trace import SlTrace

   
""" Item to display canvas panel item
"""
    
class CanvasPanelItem:
    CANV_ID = 0
    # Display types top level
    DT_NONE = None                  # Not top level
    DT_CELL = "DT_CELL"             # BrailleCell
    DT_CURSOR = "DT_CURSOR"         # Cursor
    DT_MAG_SEL = "DT_MAG_SEL"       # Magnification selection
    DT_SCAN_ITEM = "DT_SCAN_ITEM"   # Scanning item display
    
    def __init__(self, canvas_panel,
                 canv_type,
                 *args,
                 disp_type=None,
                 desc = None, **kwargs):
        """ Canvas display type
        Supports main units of display, using disp_type for top level
        items
        :canvas_panel: parent panel, in which items are displayed
        :canv_type: basic display units: rect, oval,...composite
        :disp_type: top level display units default: not top level
        :desc: descriptive text e.g. "Nf3" for chess position
        """
        self.canvas_panel = canvas_panel
        CanvasPanelItem.CANV_ID += 1
        self.canv_id = CanvasPanelItem.CANV_ID
        self.canvas_panel.items_by_id[self.canv_id] = self
        self.canv_type = canv_type
        self.disp_type = disp_type
        self.desc = desc
        self.deleted = False
        self.tags = set()
        if "tags" in kwargs:
            self.tags = set(kwargs["tags"])
        self.args = args
        self.kwargs = kwargs
        
        if self.canv_type == "create_rectangle":
            self.points = [wx_Point(args[0],args[1]),
                           wx_Point(args[2],args[3])]
        elif self.canv_type == "create_bitmap":
            self.points = [wx_Point(args[0],args[1]),   # scaling like rect
                           wx_Point(args[2],args[3])]
        elif self.canv_type == "create_oval":
            self.points = [wx_Point(args[0],args[1]),
                           wx_Point(args[2],args[3])]
        elif self.canv_type == "create_line":
            points = []
            it_args = iter(self.args)
            for arg in it_args:
                if arg is tuple:
                    x,y = arg
                else:
                    x,y = arg,next(it_args)
                pt = wx_Point(x,y)
                points.append(pt)
            self.points = points
        elif self.canv_type == "create_text":
            self.points = [wx_Point(args[0],args[1])]
        elif self.canv_type == "create_cursor":
            self.points = [wx_Point(args[0],args[1]),
                           wx_Point(args[2],args[3])]
        elif self.canv_type == "create_composite":
            self.points = []
            self.comp_parts = []        # list of composite part items
            
        else:
            raise Exception(f"draw: unrecognized type: {self.canv_type}")

    def add(self, part):
        """ Add another composite part
        :part: id/CanvasPanelItem
        """
        if type(part) == int:
            part = self.canvas_panel.items_by_id[part]
        self.points.extend(part.points)     # Accumulate points of components            
        self.comp_parts.append(part)
        
    def refresh(self):
        """ Set item to be redrawn
        """
        rect = self.bounding_rect()
        self.canvas_panel.refresh_rectangle(rect)
        
    def update(self, **kwargs):
        """ Change zero or more of items keyword values
        :**kwrgs: keyword  args to change
        """
        for kw in kwargs:
            self.kwargs[kw] = kwargs[kw]
            
    def __str__(self):
        st = "CanvasPanelItem:"
        st += f"[{self.canv_id}]"
        if self.disp_type is not None:
            st += f" self.disp_type"
        st += f"{self.canv_type}"
        if self.canvas_panel.orig_pos != (0,0):
            st += f" orig: {self.canvas_panel.orig_pos}"
        st += f" orig: {self.canvas_panel.orig_size}"
        if self.args:
            st += f" args: {self.args}"
        if self.kwargs:
            st += f" kwargs: {self.kwargs}"
        if self.points is not None:
            st += f" points: {self.points}"
        return st
        

    def draw(self, points=None, rect=None):
        """ Draw canvas item
        scaled drawings if points is not None
        
        :points: characterizing points for figure
                default: self.points (initialized points)
        :rect: limit drawing to those within this rectangle
                default: no limitation - draw item
        """
        if self.deleted:
            return      # Already deleted
        
        if points is None:
            points = self.points
            
        panel = self.canvas_panel
        self.fill = panel.color
        if "fill" in self.kwargs:
            self.fill = self.kwargs["fill"]

        self.outline = "black"        
        if "outline" in self.kwargs:
            self.outline = self.kwargs["outline"]
            
        self.width = 2
        if "width" in self.kwargs:
            self.width = self.kwargs["width"]

        if "tags" in self.kwargs:
            self.tags = self.kwargs["tags"]
        
        if self.canv_type == "create_composite":
            ret = self.create_composite_draw(points, rect=rect)
        elif self.canv_type == "create_rectangle":
            ret = self.create_rectangle_draw(points, rect=rect)
        elif self.canv_type == "create_bitmap":
            ret = self.create_bitmap_draw(points, rect=rect)
        elif self.canv_type == "create_oval":
            ret = self.create_oval_draw(points, rect=rect)
        elif self.canv_type == "create_line":
            ret = self.create_line_draw(points, rect=rect)
        elif self.canv_type == "create_text":
            ret = self.create_text_draw(points, rect=rect)
        elif self.canv_type == "create_cursor":
            ret = self.create_cursor_draw(points, rect=rect)
        else:
            raise Exception(f"draw: unrecognized type: {self.canv_type}")
        
        return ret

    def get_parts(self):
        """ Get parts (non composites), self if not composite
        :returns: list of composite parts
        """
        if self.canv_type != "create_composite":
            return [self]
        
        parts = []
        for part in self.comp_parts:
            parts.extend(part.get_parts())
        return parts

    def bounding_rect_composite(self, item):
        """ Get bounding rectangle of composite
        :item: composite type
        :returns: bounding rectangle
        """
        parts = item.get_parts()
        if len(parts) == 0:
            return wx.Rect(wx.Point(0,0), wx.Point(0,0))    # Empty
        brect  = parts[0].bounding_rect()
        ul_x, ul_y = brect.GetLeft(),brect.GetTop()
        lr_x, lr_y = brect.GetRight(),brect.GetBottom()
        
        for part in parts[1:]:
            br = part.bounding_rect()
            br_ul_x, br_ul_y = br.GetLeft(),br.GetTop()
            br_lr_x, br_lr_y = br.GetRight(),br.GetBottom()
            ul_x = min(ul_x,br_ul_x)
            ul_y = min(ul_y, br_ul_y)
            lr_x = max(lr_x,br_lr_x)
            lr_y = max(lr_y, br_lr_y)
        bbrect = wx.Rect(wx.Point(ul_x,ul_y),wx.Point(lr_x,lr_y))
        return bbrect    
            
                
    def bounding_rect(self):
        """ Get item's bounding rectangle NOT FOR COMPOSITE
        :returns: wx.Rect bounding rectangle
        """
        if self.canv_type == "create_rectangle":
            brect = wx.Rect(self.points[0], self.points[1])
        elif self.canv_type == "create_bitmap":
            brect = wx.Rect(self.points[0], self.points[1])
        elif self.canv_type == "create_oval":
            brect = wx.Rect(wx.Point(self.points[0].x,self.points[0].y),
                            wx.Point(self.points[1].x, self.points[1].y))
        elif self.canv_type == "create_line":
            brect = wx.Rect(self.points[0], self.points[1])
        elif self.canv_type == "create_text":
            if len(self.points) >= 2:
                brect = wx.Rect(self.points[0], self.points[1])
            elif len(self.points) == 1:
                brect = wx.Rect(self.points[0], self.points[0])
        elif self.canv_type == "create_cursor":
            brect = wx.Rect(wx.Point(self.points[0].x,self.points[0].y),
                            wx.Point(self.points[1].x, self.points[1].y))
        elif self.canv_type == "create_composite":
            brect = self.bounding_rect_composite(self)
        else:
            raise Exception(f"draw: unrecognized type: {self.canv_type}")
        return brect
         
    ###### create_composite    
    def create_composite_draw(self, points=None, rect=None):
        """ draw composite figure
        :points: accumulated points from all components
        :rect: rectangle if not overlapping don't draw
                default: always draw
        """
        if points is None:
            points = self.points
        if len(points) == 0:
            return      # empty list

        for part in self.comp_parts:
            part.draw(rect=rect)
         
    ###### create_rectangle    
    def create_rectangle_draw(self, points, rect=None):
        """ Simulate tkinter canvas create_rectangle drawing
        :points: points deternining figure
                default: use items points
        :rect: rectangle if not overlapping don't draw
                default: always draw
        """
        if points is None:
            points = self.points
        if len(points) < 2:
            return
        
        brect = wx.Rect(points[0], points[1])
        if rect is not None and not brect.Intersects(rect):
            return
             
        dc = wx.PaintDC(self.canvas_panel.grid_panel)
        dc.SetPen(wx.Pen(self.outline, style=wx.SOLID))
        dc.SetBrush(wx.Brush(self.fill, wx.SOLID))
        dc.DrawRectangle(brect)
        SlTrace.lg(f"DrawRect: {self.fill} {points[0]},"
                   f"  {points[1]}", "draw_rect")
        pass

         
    ###### create_bitmap    
    def create_bitmap_draw(self, points, rect=None):
        """ draw bitmap
        :points: points deternining figure
                default: use items points
        :bitmap: 
                default: always draw
        """
        if points is None:
            points = self.points
        if len(points) < 2:
            return
        
        brect = wx.Rect(points[0], points[1])
        if rect is not None and not brect.Intersects(rect):
            return
             
        dc = wx.PaintDC(self.canvas_panel.grid_panel)
        SlTrace.lg(f"DrawBitMap: {self.fill} {points[0]},"
                   f"  {points[1]}", "draw_rect")
        kwargs = self.kwargs
        c_x,c_y = points[0].x,points[0].y
        dc.DrawBitmap(kwargs["bitmap"], c_x, c_y)
        pass


    ###### create_oval
    def create_oval_draw(self, points=None, rect=None):
        """ Simulate tkinter canvas create_oval
        :points: wx.Point(x0,y0), wx.Point(x1,y1)
                default: self.points
        :rect: rectangle if not overlapping don't draw
                default: always draw
        """
        if points is None:
            points = self.points
        if len(points) < 2:
            return
        brect = wx.Rect(wx.Point(points[0].x,points[0].y), wx.Point(points[1].x,points[1].y))
        if rect is not None and not brect.Intersects(rect):
            return
        
        dc = wx.PaintDC(self.canvas_panel.grid_panel)
        dc.SetPen(wx.Pen(self.outline, style=wx.SOLID))
        dc.SetBrush(wx.Brush(self.fill, wx.SOLID))
        size=wx.Size(points[1].x-points[0].x,
                            points[1].y-points[0].y)
        dc.DrawEllipse(pt=points[0], size=size)
        SlTrace.lg(f"DrawElipse: {self.fill} {points[0]} {size}", "draw_oval")
        
    #### create_line
    def create_line_draw(self, points=None, rect=None):
        """ Implement tkinter's create_line
        :args: x1,y1,...xn,yn
        :kwargs:  width=, fill=, tags=[]
        :rect: rectangle if not overlapping don't draw TBD
                default: always draw
        """
        if points is None:
            points = self.points
        if len(points) == 0:
            return      # empty list
        
        dc = wx.PaintDC(self.canvas_panel.grid_panel)
        dc.SetPen(wx.Pen(self.fill, style=wx.SOLID, width=self.width))
        dc.DrawLines(points)

    ####### create_text
    def create_text_draw(self, points=None, rect=None):
        """ Simulate tkinter canvas create_text
        :rect: rectangle if not overlapping don't draw TBD
                default: always draw
        """
        if points is None:
            points = self.points
        if len(points) < 1:
            return
        
        
        if "font" not in self.kwargs:
            font = wx.Font(10, wx.ROMAN, wx.ITALIC, wx.NORMAL) 
        else:
            font = wx.Font(self.kwargs["font"])
            
        
        dc = wx.PaintDC(self.canvas_panel.grid_panel)
        dc.SetPen(wx.Pen(self.outline, style=wx.SOLID))
        dc.SetBrush(wx.Brush(self.fill, wx.SOLID))
        
        text = self.kwargs["text"]
        dc.SetFont(font)
        title_pt = points[0]
        dc.DrawText(text=text, pt=title_pt)
                
    ###### create_cursor
    def create_cursor_draw(self, points=None, rect=None):
        """ Create cursor - oval create_oval
        :points: wx.Point(x0,y0), wx.Point(x1,y1)
                default: self.points
        :rect: rectangle if not overlapping don't draw
                default: always draw
        """
        if points is None:
            points = self.points
        if len(points) < 2:
            return
        brect = wx.Rect(wx.Point(points[0].x,points[0].y), wx.Point(points[1].x,points[1].y))
        if rect is not None and not brect.Intersects(rect):
            return
        
        dc = wx.PaintDC(self.canvas_panel.grid_panel)
        dc.SetPen(wx.Pen(self.outline, style=wx.SOLID))
        dc.SetBrush(wx.Brush(self.fill, wx.SOLID))
        size=wx.Size(points[1].x-points[0].x,
                            points[1].y-points[0].y)
        dc.DrawEllipse(pt=points[0], size=size)
        SlTrace.lg(f"DrawCursor: {self.fill} {points[0]} {size}", "draw_cursor")
        
        
    def delete(self):
        """ delete item
        """
        if self.create_composite_draw:
            for part in self.comp_parts:
                part.delete()
        self.deleted = True
    """
    ------------------------ link to canvas_panel
    """
    
    def get_screen_loc(self, panel_loc):
        return self.canvas_panel.get_screen_loc(panel_loc)    



#wx_braille_cell_list.py 16Jan2024  crs, Author
"""Conversion of braille cell display lists
Facilitate command line specification/testing plus interprocess
communication
"""
import copy

from select_trace import SlTrace

from braille_error import BrailleError
from braille_cell import BrailleCell

class BrailleCellList:
    """ List of braille cells for display
    """
    def __init__(self, cells_or_string=None):
        """ Create cell list
        :cells_or_string: array of BrailleCell
                        or
                          dictionary by ixy of BrailleCell
                        or
                          formatted string
                default: empty
        """
        self.cells = [] # List of BrailleCell
        if cells_or_string is not  None:
            SlTrace.lg(f"\nBrailleCellList:{cells_or_string}\n", "cell_list")
            if isinstance(cells_or_string, str):
                self.cells = self.get_from_string(cells_or_string)
            elif isinstance(cells_or_string,list):
                self.cells = [self.tuple_to_braille_cell(tp)
                            for tp in cells_or_string]
            elif isinstance(cells_or_string, dict):
                for ixy in cells_or_string:
                    tp = cells_or_string[ixy]
                    self.cells.append(self.tuple_to_braille_cell(tp))
            else:
                raise BrailleError("Unsupported BrailleCellList init"
                                f" type:{type(cells_or_string)})")
        
    
    def get_cells(self):
        """ Returns list
        """
        return self.cells
        
    def get_from_string(self, string):
        """ Convert formatted string, placing result
        internally. Cells are added as found
        :string: formatted text representing a list of cells
            (ix, iy, color)*
            where:
                ix is x (horizontal index)
                iy is y (vertical index)
                color is a color string
                whitespace is ignored
        :returns: list of BrailleCell
        """
        self.gs_set_string(string)
        while True:
            bc = self.gs_get_cell()
            if bc is None:
                break   # End of list
            self.cells.append(bc)
        return self.cells
    
    def tuple_to_braille_cell(self, tp):
        """ Create BrailleCell from (ix,iy,color)
        if given BrailleCell return it unchanged
        """
        if isinstance(tp, BrailleCell):
            return tp
        
        ix,iy,color = tp
        bc = BrailleCell(ix=ix, iy=iy, color=color)
        return bc
        
    def get_from_cells(self, cells):
        """ Convert cells dictionary to list, placing result
            converts (ix,iy,color) to BrailleCell
        internally. Cells are added as found
        :cells: list/dictionary of BrailleCells/(ix,iy,color)
        :returns: updated list of BrailleCell
        """
        cell_list = cells
        if type(cells) == dict:
            cell_list = []
            for ixy in cells:
                cell = cells[ixy]
                self.cells.append(self.tp_to_braille_cell(cell))
        else:
            clist = [self.tp_to_braille_cell(tp) for tp in cell_list]
            self.cells.extend(clist)    
        return self.cells

    def get_from_text_map(self, text):
        """ Get from text map string
        :text: map using blanks, and single char colors e.g. "r" for red
        :returns: list of BrailleCell
        """
        ix = 0
        iy = 0
        
    def to_string(self, inc=False):
        """ Return parseable string
        :inc: incremental True-just show changed values
                default: False-show absolute values
        :returns: parseable string
        """
        st = ""
        prev_cell = None
        for cell in self.cells:
            ix = cell.ix 
            iy = cell.iy 
            color = cell._color
            ix_str = str(ix)
            iy_str = str(iy)
            color_str = str(cell._color)
            if inc and prev_cell is not None:
                if ix == prev_cell.ix:
                    ix_str = ""
                if iy == prev_cell.iy:
                    iy_str = ""
                if color == prev_cell._color:
                    color_str = ""
            cs = f"({ix_str},{iy_str},{color_str})"
            st += cs
            prev_cell = cell        
        return st         
        
        """
        Parsing functions
        Used to parse string to cell list
        start with "gs_"
        """
    def gs_get_cell_prev(self):
        """ Return most recently added cell
        :returns: BrailleCell added, None if none
        """
        if len(self.cells) == 0:
            return None
        
        return self.cells[-1]   # Last added
            
    def gs_set_string(self, string=""):
        """ Initialize string conversion
        :string: string to process default: empty
        """
        self.gs_string = string
        self.gs_index = 0       # Index of next character
        self.gs_end_index = len(self.gs_string)
        self.gs_cell = None     # Current cell, if any
        self.gs_new_cell()

    def gs_new_cell(self):
        """ Start new cell
        """
        self.gs_cell_prev = copy.copy(self.gs_cell)
        self.gs_cell = BrailleCell()

    def gs_error(self, string="Unknown"):
        """ Generate diagnostic error in string parsing
        """
        string  += ": "
        err_str = ( "\n" + f"{string}"
                  + "\n" + len(string)*" " + self.gs_string
                  + "\n" + len(string)*" " + self.gs_index*" " + "^"                  
                    )
        raise BrailleError(err_str)
                     
    def gs_get_cell(self):
        """ Get next cell from string, if one
        :returns: BrailleCell if one, None if no more cells
        """
        self.gs_new_cell()
        if self.gs_get_start_cell() is None:
            return None
        ix = self.gs_get_index()
        if ix is None:
            self.gs_error(f"ix missing")
        self.gs_cell.set_ix(ix)
        iy = self.gs_get_index()
        if iy is None:
            self.gs_error(f"iy missing")
        self.gs_cell.set_iy(iy)
        color = self.gs_get_color()
        if color is None or color == "":
            cell_prev = self.gs_get_cell_prev()
            if cell_prev is not None:
                color = cell_prev.color_string()
        if color is not None and color != '':
            self.gs_cell.color_cell(color)
        if self.gs_get_end_cell() is None:
            self.gs_error(f"cell end missing")
        cell = copy.copy(self.gs_cell)
        self.gs_new_cell()
        return cell

    def gs_get_index(self):
        """ Get next index
        :returtns: index, None if none
        """
        st = self.gs_get_string()
        if st == "":
            return None

        return int(st)
    
    def gs_skip_space(self):
        """ Skip past whitespace
        :returns: character after whitespace, None if end of string
        """
        while self.gs_index < self.gs_end_index:
            ch = self.gs_string[self.gs_index]
            if not ch.isspace():
                return ch
            self.gs_index += 1
        return None

    def gs_get_start_cell(self):
        """ Position after start of cell "("
        :returns: first character, None if no more string
        """
        ch = self.gs_skip_space()
        if ch != '(':
            if ch is None:
                return None # End of string
            
            self.gs_error(f"{ch} found instead of ( for cell start")    
        self.gs_index += 1
        return ch
    
    def gs_get_end_cell(self):
        """ Position at end of cell ")"
        :returns: end character, error if not found
        """
        ch = self.gs_skip_space()
        if ch != ')':
            self.gs_error(f"{ch} found instead of ) for cell end")    
        self.gs_index += 1
        return ch

    def gs_get_color(self):
        """ Get color
        Using gs_get_string needs to be changed
        if support of #.... is wanted"""
        return self.gs_get_string()
                
    def gs_get_string(self):
        """ Get string digits/or alpha
        skips leading whitespace, stops at whitespace or
        non alphanumeric char
        :returns: string
        """
        st = ""
        self.gs_skip_space()
        while self.gs_index < self.gs_end_index:
            ch = self.gs_string[self.gs_index]
            if not ch.isalnum():
                break
            st += ch
            self.gs_index += 1
        if self.gs_index < self.gs_end_index:
            ch = self.gs_string[self.gs_index]
            if ch != ")":
                self.gs_index += 1  # pass arg delimiter unless end                
        return st        
                
if __name__ == "__main__":
    
    strs = [
        "(1,1,red) (2,2) (3,3) (4,4,green) (4,3) (4,2,blue)",
        "(100,1,red) (100,2,) (100,3,) (100,4,blue) (100,5,)",
        ]
    
    def str_test(st, inc=None):
        SlTrace.lg(f"st: {st}")
        cl = BrailleCellList(st)        
        cells = cl.get_cells()
        for cell in cells:
            SlTrace.lg(str(cell))
        if inc is None:
            inct = False
            st_2 = cl.to_string(inc=inct)
            SlTrace.lg(f"Recover inc:{inct} {st_2}")
            inct = True
            st_2 = cl.to_string(inc=inct)
            SlTrace.lg(f"Recover inc:{inct} {st_2}")
        else:
            inct = True
            st_2 = cl.to_string(inc=inct)
            SlTrace.lg(f"Recover inc:{inct} {st_2}")
            
        
    for st in strs:
        str_test(st,inc=None)        